#ifndef STRUCTS_H
#define STRUCTS_H

#include <graphics.h>
#include <math.h>

/*
  A T64Poly is a polygon in '64 coordinates.  It contains all the geometry
  necessary to represent a polygon on the screen, sans the TEXTURE and
  CLIPPING, which are specific to the scenario in which the polygon is
  drawn.
*/
struct T64Poly {
  int A;
  int C;
  int X;
  int W;
  int Dir;

  T64Poly() {}
  T64Poly(int a,int c, int x, int w) {
    A = a; C = c; X = x; W = w;
  }
};
typedef T64Poly TCPoly;

/*
  A T64ViewPoly is a T64Poly plus the additional information needed to
  calculate whether it is drawn, its texture, and its clipping.
*/
struct T64ViewPoly : public T64Poly {
  int MapX;
  int MapY;
  unsigned char MapSide;
};

/*
  A T64View is a listing of T64ViewPolies for a given (SX,SY,R) that can
  be used to generate the corresponding view.
*/

#define MAXFACE 400  // max polies per view

struct T64View {
  T64Poly Poly[MAXFACE];
  int PolyBufSize;
};

/*
  A T3DEdge represents a grid point in the 3D world.  Two of these make up
  a T3DFace, which can be transformed into a T64Poly.  Four of these make up
  a T3DCube, which can be transformed into two T64ViewPolies.
*/

typedef long double wcoor;

/*
struct T3DEdge {
  wcoor flx;
  wcoor fly;

  wcoor cex;
  wcoor cey;
};

struct T3DFace {
  T3DEdge Left;
  T3DEdge Right;
};

struct T3DCube {
  T3DEdge Edge[4];
};

/***************************************************************************/
/***************************************************************************/
/***************************************************************************/
/***************************************************************************/

#define MAX(X,Y) (((X)>(Y)) ? (X) : (Y))
#define MIN(X,Y) (((X)>(Y)) ? (Y) : (X))

struct TFace {
  wcoor flx1;
  wcoor fly1;
  wcoor flx2;
  wcoor fly2;

  wcoor cex1;
  wcoor cey1;
  wcoor cex2;
  wcoor cey2;

  int MapX;
  int MapY;
  int MapSide;

  int TwoIsBigger;

  TFace() {
    flx1 = 0;
    fly1 = 0;
    flx2 = 0;
    fly2 = 0;

    cex1 = 0;
    cey1 = 0;
    cex2 = 0;
    cey2 = 0;

    MapX = 0;
    MapY = 0;
    MapSide = 0;
    TwoIsBigger = 0;
  }

  wcoor A() {
    if (TwoIsBigger) return fabsl(cey2-fly2)/2.0;
    return fabsl(cey1-fly1)/2.0;
  }
  wcoor B() {
    if (!TwoIsBigger) return fabsl(cey2-fly2)/2.0;
    return fabsl(cey1-fly1)/2.0;
  }
  wcoor rightx() {
    return MAX(flx1,flx2);
  }
  wcoor leftx() {
    return MIN(flx1,flx2);
  }

  void CalcLeftRight() {
    TwoIsBigger = 0;
    if (A() > B()) return;
    TwoIsBigger = 1;
  }

  void SetMapInfo(int _MapX,int _MapY,int _MapSide) {
    MapX = _MapX;
    MapY = _MapY;
    MapSide = _MapSide;
  }

  TFace& operator=(TFace& src) {
    flx1 = src.flx1;
    fly1 = src.fly1;
    flx2 = src.flx2;
    fly2 = src.fly2;

    cex1 = src.cex1;
    cey1 = src.cey1;
    cex2 = src.cex2;
    cey2 = src.cey2;

    MapX = src.MapX;
    MapY = src.MapY;
    MapSide = src.MapSide;

    TwoIsBigger = src.TwoIsBigger;
    return *this;
  }
};

struct TC64Poly {
  int A;
  int B;
  int C;
  int W;
  int X;
  int Dir;

  int MapX;
  int MapY;
  int MapSide;

  TC64Poly() {}
  TC64Poly(const TC64Poly& src) { *this = src; }

  TC64Poly& operator=(const TC64Poly& src) {
    A = src.A;
    B = src.B;
    C = src.C;
    W = src.W;
    X = src.X;
    Dir = src.Dir;
    MapX = src.MapX;
    MapY = src.MapY;
    MapSide = src.MapSide;
    return *this;
  }

  void Face2C64(TFace face);

  int leftx() {
    if (Dir == 0) return X - W; // A is to right of B
    else return X;
  }
  int rightx() {
    if (Dir == 0) return X; // A is to right of B
    else return X + W;
  }
};

char GetMap(int X,int Y);

extern char mymap[32][32];

extern TFace FaceBuf[MAXFACE];
extern unsigned FaceBufSize;

extern TC64Poly PolyBuf[MAXFACE];
extern unsigned PolyBufSize;

extern char AttrTable[][24];
extern char ImgTable[][24];

#endif
