#include "structs.h"
#include "v64.h"
#include "video.h"

#include <conio.h>
#include <process.h>
#include <fstream.h>
#include <stdio.h>
#include <dos.h>

int clipping = 1;
int rounding = 1;  // 1=roundit, 0=don't

void C64MoveTo(int x,int y) {
  x *= V64_PW;
  x += V64_ULX;
  y *= V64_PH;
  y += V64_ULY;

  moveto(x+V64_PW/2,y+V64_PH/2);
}

void C64LineTo(int x,int y) {
  x *= V64_PW;
  x += V64_ULX;
  y *= V64_PH;
  y += V64_ULY;

  lineto(x+V64_PW/2,y+V64_PH/2);
}


char checkerboard[8] = {
  0xAA,   /* 10101010  =        */
  0x55,   /* 01010101  =        */
  0xAA,   /* 10101010  =        */
  0x55,   /* 01010101  =        */
  0xAA,   /* 10101010  =        */
  0x55,   /* 01010101  =        */
  0xAA,   /* 10101010  =        */
  0x55    /* 01010101  =        */
};


void C64PutPixel(int x,int y,int color) {
  if ((x < 0) || (x >= V64_W) || (y < 0) || (y >= V64_H))
    setfillpattern(checkerboard,color);
  else setfillstyle(SOLID_FILL,color);

  x *= V64_PW;
  x += V64_ULX;
  y *= V64_PH;
  y += V64_ULY;

  bar(x+1,y+1,x+V64_PW-1,y+V64_PH-1);
}

int C64GetPixel(int x,int y) {
  if ((x < 0) || (x >= V64_W) || (y < 0) || (y >= V64_H)) return 0;

  x *= V64_PW;
  x += V64_ULX;
  y *= V64_PH;
  y += V64_ULY;

  return getpixel(x+1,y+1);
}

void C64ClearScr() {
  // first, draw the border around it:
  setcolor(WHITE);

  int ulx = V64_ULX;
  int uly = V64_ULY;
  int lrx = V64_ULX + (V64_W) * V64_PW;
  int lry = V64_ULY + (V64_H) * V64_PH;
  int x,y,i;

  for (i = 0; i < 5; i++)
    rectangle(ulx-i,uly-i,lrx+i+1,lry+i+1);

  for (x = 0; x < V64_W; x++)
    for (y = 0; y < V64_H; y++)
      C64PutPixel(x,y,BLACK);

  setcolor(DARKGRAY);

  // now draw gridlines:
  for (x = 0; x <= V64_W; x++) {
    if (x == (V64_W / 2)+1) setcolor(LIGHTGRAY);
    line(ulx + (x*V64_PW),uly,ulx + (x*V64_PW),lry);
    if (x == (V64_W / 2)+1) setcolor(DARKGRAY);
  }
  for (y = 0; y <= V64_H; y++) {
    if (y == (V64_H / 2)+1) setcolor(LIGHTGRAY);
    line(ulx,uly + (y*V64_PH),lrx,uly + (y*V64_PH));
    if (y == (V64_H / 2)+1) setcolor(DARKGRAY);
  }
}


void DrawPolyAL(TC64Poly Poly,int Clip,int Color) {
  int cc;  // columncount

  for (cc=Clip; cc<=Poly.W; cc++) {
    if (clipping)
      if (C64GetPixel(Poly.X - cc,V64_MIDDLE)) break;

    int y;
    if (Poly.W) y = (Poly.A - (Poly.C * cc / (float) Poly.W)) + (rounding/2.0);
    else y = Poly.A;

    for (int hdiv2=0; hdiv2<y; hdiv2++) {
      C64PutPixel(Poly.X - cc,V64_MIDDLE - hdiv2,Color);
      C64PutPixel(Poly.X - cc,V64_MIDDLE + hdiv2 + 1,Color);
    }
  }
}

void DrawPolyAR(TC64Poly Poly,int Clip,int Color) {
  int cc;  // columncount

  for (cc=Poly.W-Clip; cc>=0; cc--) {
    if (clipping)
      if (C64GetPixel(Poly.X - cc,V64_MIDDLE)) break;

    int y;
    if (Poly.W) y = (Poly.A - (Poly.C * cc / (float) Poly.W)) + (rounding/2.0);
    else y = Poly.A;

    for (int hdiv2=0; hdiv2<y; hdiv2++) {
      C64PutPixel(Poly.X - cc,V64_MIDDLE - hdiv2,Color);
      C64PutPixel(Poly.X - cc,V64_MIDDLE + hdiv2 + 1,Color);
    }

  }
}

void DrawPolyBL(TC64Poly Poly,int Clip,int Color) {
  int cc;  // columncount

  for (cc=Poly.W-Clip; cc>=0; cc--) {
    if (clipping)
      if (C64GetPixel(Poly.X + cc,V64_MIDDLE)) break;

    int y;
    if (Poly.W) y = (Poly.A - (Poly.C * cc / (float) Poly.W)) + (rounding/2.0);
    else y = Poly.A;

    for (int hdiv2=0; hdiv2<y; hdiv2++) {
      C64PutPixel(Poly.X + cc,V64_MIDDLE - hdiv2,Color);
      C64PutPixel(Poly.X + cc,V64_MIDDLE + hdiv2 + 1,Color);
    }
  }
}

void DrawPolyBR(TC64Poly Poly,int Clip,int Color) {
  int cc;  // columncount

  for (cc=Clip; cc<=Poly.W; cc++) {
    if (clipping)
      if (C64GetPixel(Poly.X + cc,V64_MIDDLE)) break;

    int y;
    if (Poly.W) y = (Poly.A - (Poly.C * cc / (float) Poly.W)) + (rounding/2.0);
    else y = Poly.A;

    for (int hdiv2=0; hdiv2<y; hdiv2++) {
      C64PutPixel(Poly.X + cc,V64_MIDDLE - hdiv2,Color);
      C64PutPixel(Poly.X + cc,V64_MIDDLE + hdiv2 + 1,Color);
    }
  }
}

void _DrawItNormal(TC64Poly Poly,int Color) {
  if (!clipping)  {
    if (Poly.Dir == 0) DrawPolyAR(Poly,0,Color);
    else DrawPolyBR(Poly,0,Color);
  }

  if (Poly.Dir == 0) {
    // ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; TYPE A ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    // Case 1 - B is off the left border

    if (Poly.X - Poly.W < 0) {
      DrawPolyAR(Poly,Poly.W-Poly.X,Color);
      return;
    }

    // Case 2 - A is off the right border

    if (Poly.X >= V64_W) {
      DrawPolyAL(Poly,Poly.X - (V64_W-1),Color);
      return;
    }

    // Case 3 - A is hidden, but B isn't
    if (C64GetPixel(Poly.X,V64_MIDDLE)) {
      // give up if both A and B are hidden
      if (C64GetPixel(Poly.X-Poly.W,V64_MIDDLE)) return;

      DrawPolyAR(Poly,0,Color);
      return;
    }

    // Case 4 - A is not hidden, possibly B is

    DrawPolyAL(Poly,0,Color);
    return;
  }


  // ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; TYPE B ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

  // Case 1 - A is off the left border

  if (Poly.X < 0) {
    DrawPolyBR(Poly,-Poly.X,Color);
    return;
  }

  // Case 2 - B is off the right border

  if (Poly.X + Poly.W >= V64_W) {
    DrawPolyBL(Poly,Poly.X + Poly.W - (V64_W-1),Color);
    return;
  }

  // Case 3 - A is hidden, but B isn't

  if (C64GetPixel(Poly.X,V64_MIDDLE)) {
    // give up if both A and B are hidden:
    if (C64GetPixel(Poly.X + Poly.W,V64_MIDDLE)) return;

    DrawPolyBL(Poly,0,Color);
    return;
  }

  // Case 4 - A is not hidden, possibly B is
  DrawPolyBR(Poly,0,Color);
}

void DrawCorners(TC64Poly Poly,int Color) {
  setcolor((Color+16)%16);
  if (!Poly.Dir) setlinestyle(DASHED_LINE,0,1);


  if (Poly.Dir) {
/*
    C64PutPixel(Poly.X,V64_MIDDLE - Poly.A + 1,Color+1);
    C64PutPixel(Poly.X,V64_MIDDLE + Poly.A,Color+1);
    C64PutPixel(Poly.X+Poly.W,V64_MIDDLE - (Poly.A-Poly.C) + 1,Color+1);
    C64PutPixel(Poly.X+Poly.W,V64_MIDDLE + (Poly.A-Poly.C),Color+1);
/*
    C64MoveTo(Poly.X,V64_MIDDLE - Poly.A + 1);
    C64LineTo(Poly.X+Poly.W,V64_MIDDLE + (Poly.A-Poly.C));
    C64MoveTo(Poly.X,V64_MIDDLE + Poly.A);
    C64LineTo(Poly.X+Poly.W,V64_MIDDLE - (Poly.A-Poly.C) + 1);
*/
    C64MoveTo(Poly.X,V64_MIDDLE - Poly.A + 1);
    C64LineTo(Poly.X,V64_MIDDLE + Poly.A);
    C64LineTo(Poly.X+Poly.W,V64_MIDDLE + (Poly.A-Poly.C));
    C64LineTo(Poly.X+Poly.W,V64_MIDDLE - (Poly.A-Poly.C) + 1);
    C64LineTo(Poly.X,V64_MIDDLE - Poly.A + 1);
  } else {
/*
    C64PutPixel(Poly.X,V64_MIDDLE - Poly.A + 1,Color+1);
    C64PutPixel(Poly.X,V64_MIDDLE + Poly.A,Color+1);
    C64PutPixel(Poly.X-Poly.W,V64_MIDDLE - (Poly.A-Poly.C) + 1,Color+1);
    C64PutPixel(Poly.X-Poly.W,V64_MIDDLE + (Poly.A-Poly.C),Color+1);

/*
    C64MoveTo(Poly.X,V64_MIDDLE - Poly.A + 1);
    C64LineTo(Poly.X-Poly.W,V64_MIDDLE + (Poly.A-Poly.C));
    C64MoveTo(Poly.X,V64_MIDDLE + Poly.A);
    C64LineTo(Poly.X-Poly.W,V64_MIDDLE - (Poly.A-Poly.C) + 1);
*/
    C64MoveTo(Poly.X,V64_MIDDLE - Poly.A + 1);
    C64LineTo(Poly.X,V64_MIDDLE + Poly.A);
    C64LineTo(Poly.X-Poly.W,V64_MIDDLE + (Poly.A-Poly.C));
    C64LineTo(Poly.X-Poly.W,V64_MIDDLE - (Poly.A-Poly.C) + 1);
    C64LineTo(Poly.X,V64_MIDDLE - Poly.A + 1);
  }

  if (!Poly.Dir) setlinestyle(SOLID_LINE,0,1);
}

void DrawFloorGrid(TC64Poly Poly,int Color) {
  if (!Poly.W) return;
  if (Poly.Dir) {
    //C64PutPixel(Poly.X,V64_MIDDLE + Poly.A,Color);
    //C64PutPixel(Poly.X+Poly.W,V64_MIDDLE + (Poly.A-Poly.C),Color);

    for (int x=0;x<=Poly.W;x++) {
      int y = (x * -Poly.C / (float)Poly.W) + Poly.A + (rounding/2.0);

      if (Poly.X+x < 0 || Poly.X+x >= V64_W) continue;
      if (V64_MIDDLE + y >= V64_H) continue;
      C64PutPixel(Poly.X+x,V64_MIDDLE + y,Color);
    }

  } else {
    //C64PutPixel(Poly.X,V64_MIDDLE + Poly.A,Color+1);
    //C64PutPixel(Poly.X-Poly.W,V64_MIDDLE + (Poly.A-Poly.C),Color+1);

    for (int x=0;x<=Poly.W;x++) {
      int y = (x * -Poly.C / (float)Poly.W) + Poly.A + (rounding/2.0);

      if (Poly.X-x < 0 || Poly.X-x >= V64_W) continue;
      if (V64_MIDDLE + y >= V64_H) continue;
      C64PutPixel(Poly.X-x,V64_MIDDLE + y,Color);
    }
  }
}


void DrawItNormal(TC64Poly Poly,int Color) {
  if (ViewAttr & 1) {
    // flip it
    Poly.X = (V64_W-1)-Poly.X;
    Poly.Dir = !Poly.Dir;
  }

  static c = 1;
  c++;
  if (c == 15) c = 1;

  if (Color == -1) _DrawItNormal(Poly,c);
  else _DrawItNormal(Poly,Color);

  if (WireFrame) DrawCorners(Poly,Color+8);
}
