/*
  ALIAS.CPP - Written by Peter Gonzalez [pbgonz@mail.wm.edu], 1995

    This file defines functions for reading an ALIAS.  It is called by
  ReadAliasesAndBlocks() in PARSER.CPP.

    CASM, CDASM, CLINK, and CDUMP are freeware; they may be freely
  distributed, provided that no fee is charged and I am given due
  credit for my work.  This source code is available directly from me upon
  request; please do not distribute the source code or any compiled
  modifications without my consent.  Support freeware!
*/

#include "alias.h"
#include "block.h"
#include "numbers.h"
#include "casm.h"

THash AliasList;

TAutoAlias *AutoAliasesB=0;
TAutoAlias *AutoAliasesW=0;

//--------------------------------------------------------------------------
void ReadVARS(TASMFileTokenizer& ftok) {
//--------------------------------------------------------------------------
  TToken cur;

  cur = ftok.PeekNextToken();	// no vars found
  if (cur != "VARS") return;
  ftok.GetNextToken();

  TAutoAlias *LastB=0;
  TAutoAlias *LastW=0;

  BOOL OpenRange = FALSE;

  while (1) {
    cur = ftok.PeekNextToken();
    if (cur == "ENDV") {
      ftok.GetNextToken();
      if (OpenRange) {
	cur.WriteFileLine(ErrStr);
	ErrStr << "Expecting address, found \"ENDV\"" << ENDE;
      }
      break;
    }

    unsigned value;
    int result = ReadByteOrWord(ftok,value);
    if (!result) {
      cur.WriteFileLine(ErrStr);
      ErrStr << "Expecting address, found \"" << cur.Text << '"' << ENDE;
    }

    TAutoAlias *newa = new TAutoAlias;
    if (!newa) ErrStr.OutOfMemory();

    newa->Addr = value;

    newa->IsRange = FALSE;
    if (!OpenRange) {
      if (ftok.PeekNextToken() == '-') {
	ftok.GetNextToken();
	newa->IsRange = TRUE;
	OpenRange = TRUE;
      }
    } else OpenRange = FALSE;

    // push it on the end:
    newa->Next = 0;
    if (result < 0) {  // byte
      if (!LastB) AutoAliasesB = newa;
      else LastB->Next = newa;
      LastB = newa;
    } else {  // word
      if (!LastW) AutoAliasesW = newa;
      else LastW->Next = newa;
      LastW = newa;
    }
  }
}

//--------------------------------------------------------------------------
TAlias *ReadAliasDef(TASMFileTokenizer& ftok,BOOL IsConst,BOOL IsWord,
  TBlock *LocalBlock) {
//--------------------------------------------------------------------------
  TToken cur = ftok.GetNextToken();
  if (!IsID(cur.Text)) {
    cur.WriteFileLine(ErrStr);
    ErrStr << "Invalid characters in identifier" << ENDE;
  }

  if (IsNameInUse(cur.Text)) {
    cur.WriteFileLine(ErrStr);
    ErrStr << "The identifier \"" << cur.Text << "\" is already in use" << ENDE;
  }

  if (ftok.GetNextToken() != '=') {
    cur.WriteFileLine(ErrStr);
    ErrStr << "'=' expected" << ENDE;
  }

  TAlias *NewAlias = new TAlias;
  if (!NewAlias) ErrStr.OutOfMemory();
  NewAlias->Name = cur.Text;
  NewAlias->Attrs.IsUsed = FALSE;
  NewAlias->Attrs.IsConst = IsConst;

  BOOL SubtractIt = FALSE;
  unsigned newval;
  unsigned Param = 0;
  enum { NONE, HI, LO } ArrowRead;  // "", ">", "<"

  while(1) {
    cur = ftok.PeekNextToken();

    if (cur == '>') {
      ftok.GetNextToken(); // extract '>'
      ArrowRead = HI;
      cur = ftok.PeekNextToken();
    } else if (cur == '<') {
      ftok.GetNextToken(); // extract '<'
      ArrowRead = LO;
      cur = ftok.PeekNextToken();
    } else ArrowRead = NONE;

    switch (ReadByteOrWord(ftok,newval)) {
      case 0: // nothing read
	// perhaps it's an alias
	if (isidchar(cur.Text[0])) {
	  // check for an alias:
	  TAlias *CurAlias=0;
	  if (LocalBlock)
	    CurAlias = (TAlias *)LocalBlock->LocalAliases.Find(cur.Text.c_str());
	  if (!CurAlias)
	    CurAlias = (TAlias *)AliasList.Find(cur.Text.c_str());

	  if (CurAlias) {
	    ftok.GetNextToken(); // read the peeked text

	    CurAlias->Attrs.IsUsed = TRUE;  // for "unused alias" error

	    newval = CurAlias->Data;
	  } else {
	    cur.WriteFileLine(ErrStr);
	    ErrStr << "Unknown alias : \""
	      << cur << '"' << ENDE;
	  }
	} else {
	  cur.WriteFileLine(ErrStr);
	  ErrStr << "Syntax error: \"" << cur << '"' << ENDE;
	}
	break;
      case 1:  // word
      case -1: // byte
	// keep newval as it is
	break;
    }

    if (ArrowRead == HI) {
      newval = newval >> 8;
    } else if (ArrowRead == LO) {
      newval = newval & 0xff;
    }

    if (SubtractIt) Param -= newval;
    else Param += newval;

    cur = ftok.PeekNextToken();
    if (cur == '-') {
      ftok.GetNextToken();  // read the '-'
      SubtractIt = TRUE;
      continue;
    } else if (cur == '+') {
      ftok.GetNextToken();  // read the '+'
      SubtractIt = FALSE;
      continue;
    }

    // otherwise, quit the loop
    break;
  }

  NewAlias->Data = Param;
  if (IsWord) {
    NewAlias->Attrs.IsWord = TRUE;
  } else {
    NewAlias->Attrs.IsWord = FALSE;
    if (Param & 0xff00) {
      cur.WriteFileLine(ErrStr);
      ErrStr << "Expecting 8-bit value, not 16-bit value" << ENDE;
    }
  }
  return NewAlias;
}

//--------------------------------------------------------------------------
TAlias *ReadAutoAliasDef(TASMFileTokenizer& ftok,BOOL IsWord) {
//--------------------------------------------------------------------------
  TToken cur = ftok.GetNextToken();
  if (!IsID(cur.Text)) {
    cur.WriteFileLine(ErrStr);
    ErrStr << "Invalid characters in identifier" << ENDE;
  }

  if (IsNameInUse(cur.Text)) {
    cur.WriteFileLine(ErrStr);
    ErrStr << "The identifier \"" << cur.Text << "\" is already in use" << ENDE;
  }

  TAlias *NewAlias = new TAlias;
  if (!NewAlias) ErrStr.OutOfMemory();
  NewAlias->Name = cur.Text;
  NewAlias->Attrs.IsUsed = FALSE;
  NewAlias->Attrs.IsConst = FALSE;
  NewAlias->Attrs.IsWord = IsWord;


  // now calculate the address:

  TAutoAlias **List;
  if (IsWord) List = &AutoAliasesW;
  else List = &AutoAliasesB;

  if (!*List) {
    cur.WriteFileLine(ErrStr);
    ErrStr << "Ran out of vars" << ENDE;
  }

  NewAlias->Data = (*List)->Addr;
  TAutoAlias *del;

  if (!(*List)->IsRange) {
    del = (*List);
    (*List) = (*List)->Next;
    delete del;
  } else {  // it's a range
    if ((*List)->Next->Addr > (*List)->Addr) {
      // increment it
      (*List)->Addr++;
    } else {
      // remove both:
      del = (*List);
      (*List) = (*List)->Next;
      delete del;
      if (!(*List)) ErrStr << "Missing range terminator" << PROGBUG;
      del = (*List);
      (*List) = (*List)->Next;
      delete del;
    }
  }

  return NewAlias;
}

// Shared by multiple instances of recursive routines below:
static BOOL AnyUnusedAliases;
static char *BlockName;

//--------------------------------------------------------------------------
void _ShowUnusedAliases(TAlias *Alias) {
//--------------------------------------------------------------------------
  if (!Alias) return;

  if (!Alias->Attrs.IsUsed && !Alias->Attrs.IsConst) {
    if (!AnyUnusedAliases) {
      AnyUnusedAliases = TRUE;
      ErrStr << "The following aliases are defined, but not used:\n" << ENDW;
    }
    cout << "  ";
    if (BlockName) cout << BlockName << "|";
    cout << Alias->Name << '\n';
  }

  for (int i=0; i<BTREENUMBRANCHES; i++)
    _ShowUnusedAliases((TAlias *)Alias->Branch[i]);
}


//--------------------------------------------------------------------------
void _ShowBlockAliases(TBlock *Block) {
//--------------------------------------------------------------------------
  if (!Block) return;

  BlockName = (char *)Block->Name.c_str();
  _ShowUnusedAliases((TAlias *)Block->LocalAliases.GetRoot());

  for (int i=0; i<BTREENUMBRANCHES; i++)
    _ShowBlockAliases((TBlock *)Block->Branch[i]);
}

//--------------------------------------------------------------------------
void ShowUnusedAliases() {
//--------------------------------------------------------------------------

  AnyUnusedAliases = FALSE;

  // First, do the global aliases:
  BlockName = 0;
  _ShowUnusedAliases((TAlias *)AliasList.GetRoot());

  // Now do the local aliases:
  _ShowBlockAliases((TBlock *)BlockList.GetRoot());
}

