/* 
  BLOCK.H - Written by Peter Gonzalez [pbgonz@mail.wm.edu], 1995 
 
    CASM, CDASM, CLINK, and CDUMP are freeware; they may be freely 
  distributed, provided that no fee is charged and I am given due 
  credit for my work.  This source code is available directly from me upon 
  request; please do not distribute the source code or any compiled  
  modifications without my consent.  Support freeware!  
*/ 
 
#ifndef _BLOCK_H
#define _BLOCK_H

#include "hash.h"
#include "petestr.h"
#include "tokenize.h"

struct TBlock : public THashObject {
  enum TBlockType { CODE, DATA, UNDEFINED } BlockType;

  BOOL DefinedYet;
  char *ObjCode; // pointer to assembled code
  unsigned ObjCodeSize; // size of data pointed to by ObjCode

  unsigned OffsetInDestFile; // location in output file

  // Aliases whose scope is only within this block
  THash LocalAliases;

  // label definitions (each block has itself as a definition)
  struct TLabelDef : public THashObject {
    unsigned OffsetInBlock;
    TLabelDef(const FString& _LabelName,unsigned _OffsetInBlock = 0) {
      Name = _LabelName;
      OffsetInBlock = _OffsetInBlock;
    }
  };
  THash LabelDefs;

  // label references
  struct TLabelRef : public THashObject {
    TBlock *DefBlock;       // The block in which the label should be defined
    unsigned OffsetInBlock; // Where in this block the label is referenced

    unsigned AddAmount;     // this gets added to the address being stored
			    // (can be a two's complement)

    FString Filename;
    long LineNum; // used for error handling

    enum TRefType {
      HI,    // LDA #>Label    (add the high byte of the label's addr)
      LO,    // LDA #<Label    (add the low byte of the label's addr)
      NEAR,  // BEQ            (add the two's complement offset of the label)
      FAR    // JMP            (add the full 16-bit label)
    } RefType;
  };
  THash LabelRefs;

  TBlock(const char *name);
  ~TBlock() {  }
};
extern THash BlockList;

void ReadBlock(TASMFileTokenizer& ftok,TBlock::TBlockType InitialMode);

struct TNearJump : public THashObject {
  TBlock *ParentBlock;
  unsigned StartOffset;  // which byte to change in ParentBlock's code
  unsigned EndAddr;
  FString Filename;
  long LineNum;

  TNearJump(TBlock *_ParentBlock,unsigned _StartOffset,unsigned _EndAddr,
    const FString& _Filename,long _LineNum) {
    ParentBlock = _ParentBlock;
    StartOffset = _StartOffset;
    EndAddr = _EndAddr;
    Filename = _Filename;
    LineNum = _LineNum;
  }
};
extern THash NearJumpList;


#endif
