/*
  CASM.CPP - Written by Peter Gonzalez [pbgonz@mail.wm.edu], 1995

    CASM, CDASM, CLINK, and CDUMP are freeware; they may be freely
  distributed, provided that no fee is charged and I am given due
  credit for my work.  This source code is available directly from me upon
  request; please do not distribute the source code or any compiled
  modifications without my consent.  Support freeware!
*/

#ifndef CASM
#error The macro definition "CASM" must be defined when building CASM.
#endif

#include <iostream.h>
#include <alloc.h>
#include <malloc.h>
#include <string.h>
#include <dos.h>
#include <sys\timeb.H>
#include <stdio.h>

#include "casm.h"
#include "block.h"
#include "tables.h"
#include "diskacc.h"
#include "version.h"
#include "layout.h"
#include "parser.h"
#include "alias.h"
#include "showdump.h"
#include "fixup.h"
#include "numbers.h"

unsigned _stklen = 0x2000;  // double the stack size to 8K

//--------------------------------------------------------------------------
TGlobalParams::TGlobalParams() {
//--------------------------------------------------------------------------
  LeaveOutLoadAddress = FALSE;
  ShowDisassembly = FALSE;
  HideWarnings = FALSE;
  QuietMode = FALSE;
  ProgStart = 0xffff;

  OBufSize = 0;
  OBuf = (char huge *) farmalloc(0xff01);
  if (!OBuf) ErrStr.OutOfMemory();

  ObjBufSize = 0;
  ObjBuf = (char huge *) farmalloc(0xff01);

  if (!ObjBuf) ErrStr.OutOfMemory();

  memset(OBuf,0,0xff00);
  memset(ObjBuf,0,0xff00);
}

TGlobalParams GP;

//--------------------------------------------------------------------------
void WriteOutputFile() {
//--------------------------------------------------------------------------
  DString OutFileName = GP.MainFileName;
  OutFileName.RemoveFileExtension();
  OutFileName += ".PRG";

  unsigned OutFile = fasdos_create(OutFileName.c_str(),0);
  if (OutFile == 0xffff) {
    ErrStr << "Unable to open output file \"" << OutFileName << '\"' << ENDE;
  }

  if (!GP.LeaveOutLoadAddress)
    fasdos_write(OutFile,(char *) &GP.ProgStart,2);

  char huge *curchunk = GP.OBuf;

  while (GP.OBufSize > 0) {
    unsigned blocksize;
    if (GP.OBufSize < 0x4000) blocksize = GP.OBufSize;
    else blocksize = 0x4000;
    if (fasdos_write(OutFile,curchunk,blocksize) == 0xffff) {
      fasdos_close(OutFile);
      ErrStr << "Error writing to output file \""
	<< OutFileName << '\"' << ENDE;
    }
    curchunk += blocksize;
    GP.OBufSize -= blocksize;
  }

  fasdos_close(OutFile);
  cout.flush();

  con << '\n';
  if (GP.QuietMode) con << "      ";  // line up output with "CASM:"
  con << "File successfully assembled; "
       << OutFileName << " written to disk.\n";
  con.flush();
}

//--------------------------------------------------------------------------
double timeb2double(const timeb& time) {
//--------------------------------------------------------------------------
  double result = time.time;
  result += time.millitm/1000.0;
  return result;
}

//--------------------------------------------------------------------------
void ShowStats(timeb& StartTime,unsigned long StartMem,
  unsigned long TotalLines) {
//--------------------------------------------------------------------------
  timeb now;
  ftime(&now);
  char buf[20];
  int blocks;

  cout << "\nInput size:     " << TotalLines << " lines";

  sprintf(buf,"%.2f",timeb2double(now)-timeb2double(StartTime));
  cout << "\nAssemble time:  " << buf << " seconds";

  cout << "\nInitial Memory: " << StartMem << " bytes";

  cout << "\nMemory used:    " << StartMem - farcoreleft() << " bytes (";
  sprintf(buf,"%.1f",100.0-(100.0*farcoreleft()/(double)StartMem) );
  cout << buf << "%)";

  cout << "\nLoad Addr:      $";
  WriteHexWord(cout,GP.ProgStart);

  if (GP.LeaveOutLoadAddress) cout << " (Not stored in file)";

  cout << "\nOutput size:    $";
  WriteHexWord(cout,GP.ObjBufSize);
  blocks = (GP.ObjBufSize / 254)+1;
  cout << " bytes (" << blocks << " block"
       << ((blocks != 1) ? "s)\n" : ")\n");
}

//--------------------------------------------------------------------------
void ShowHelp() {
//--------------------------------------------------------------------------
  cout.flush();

  con << "\nCASM - Commodore 64 Assembler"
       << "\nVersion " << VersionID() << ", " << __DATE__
       << "\nBy ";

  if (!WRITENAME(con)) con << "?><FDDSEFD<";

  con << "\
\n\nUSAGE: CASM <SourceFile> [/D] [/N] [/Q] [/W]\n\
       CASM [/F] [/?]\n\
\n\
CASM assembles a *.ASM file, writing the output to a corresponding\n\
*.PRG file.  The following command-line options are available:\n\
\n\
/D  Output a disassembly (for debugging)\n\
/N  Prevent the load address from being stored at the beginning of the file\n\
/Q  Run CASM in \"quiet\" mode, where less text is displayed on the screen\n\
/W  Warning messages are not displayed\n\
/F  Displays freeware distribution information\n\
\n\
Free memory: " << farcoreleft() << " bytes\n";

  con.flush();
}

int c_break() {
  cout.flush();
  con << "\n[CTRL-C Pressed]\n";
  return 0; // return to DOS
}

void (*_FixNearJumps)() = FixNearJumps;
void writeshortbanner() {
  cout.flush();
  con << "\nCASM Version " << VersionID() << ", " << __DATE__ << ", ";
  if (!WRITENAME(con)) {
    _FixNearJumps = (void(*)()) (void *) WriteOutputFile;
  }
  con << "\n";
  con.flush();
}

//--------------------------------------------------------------------------
int main(int argc,char *argv[]) {
//--------------------------------------------------------------------------
  timeb StartTime;  // for calculating assemble time
  ftime(&StartTime);
  unsigned long StartMem = farcoreleft();

  ctrlbrk(c_break); setcbrk(1);

  int c=argc;

  while (c > 2) {
    strupr(argv[c-1]);
    if (fastcmp("/D",argv[c-1])) GP.ShowDisassembly = TRUE;
    else if (fastcmp("/N",argv[c-1])) GP.LeaveOutLoadAddress = TRUE;
    else if (fastcmp("/Q",argv[c-1])) GP.QuietMode = TRUE;
    else if (fastcmp("/W",argv[c-1])) GP.HideWarnings = TRUE;
    else if (fastcmp("/F",argv[c-1])) {
fw:   writeshortbanner();
      WriteFreewareMsg();
      return 1;
    } else { ShowHelp(); return 1; }
    c--;
  }

  if (fastcmp(argv[1],"/F")) goto fw;

  if ((c != 2) || fastcmp(argv[1],"/?")) {
    ShowHelp();
    return 1;
  }

  if (!GP.QuietMode) writeshortbanner();

  strupr(argv[1]);
  GP.MainFileName = argv[1];

  if (*argv[1] == '/')
    ErrStr << "Missing filename" << ENDE;

  if (ifstream(argv[1]).fail()) GP.MainFileName += ".ASM";
  if (ifstream(GP.MainFileName.c_str()).fail())
    ErrStr << "Unable to open \"" << argv[1] << "\"" << ENDE;

  if (GP.QuietMode) con << "CASM: ";
  else cout << '\n';

  con << "Now assembling " << GP.MainFileName << "...";

  con.flush();

  unsigned long TotalLines;
  {
    TASMFileTokenizer ftok(GP.MainFileName);
    PreScanLayoutTable(ftok);
    ReadVARS(ftok);
    ReadAliasesAndBlocks(ftok);
    TotalLines = ftok.GetTotalLines();
  } // see that the file is closed before next line is executed:

  ReScanLayoutTable(GP.MainFileName.c_str());

  ResolveLabels();
  _FixNearJumps();

  WriteOutputFile();

  con.flush();

  if (!GP.HideWarnings) ShowUnusedAliases();

  if (!GP.QuietMode) {
    cout.flush();
    ShowStats(StartTime,StartMem,TotalLines);
  }

  if (GP.ShowDisassembly) ShowDump();

  return 0;
}


