/*
  CDASM.CPP - Written by Peter Gonzalez [pbgonz@mail.wm.edu], 1995 
 
    CASM, CDASM, CLINK, and CDUMP are freeware; they may be freely 
  distributed, provided that no fee is charged and I am given due 
  credit for my work.  This source code is available directly from me upon 
  request; please do not distribute the source code or any compiled  
  modifications without my consent.  Support freeware!  
*/ 
 
#include <fstream.h>
#include <constrea.h>
#include <dos.h>

#include "tables.h"
#include "diskacc.h"
#include "version.h"
#include "numbers.h"
#include "petestr.h"

#define MAXLABELS 4096
unsigned LabelAddrs[MAXLABELS+2];
char LabelOffset[MAXLABELS+2];  // i.e. does instruction start here?
int NumLabelAddrs = 0;

char huge *IBuf;
unsigned IBufSize;
unsigned CurByte;  // offset into IBuf
unsigned LoadAddr = 0;

int FindLabel(unsigned Addr) {
  int i = 0;
  while (i < NumLabelAddrs) {
    if (LabelAddrs[i] == Addr) return i;
    i++;
  }
  return -1;
}

unsigned GetLastValidAddr(unsigned Addr) {
// checks to see if an instruction starts at Addr.  If not, it returns the
// last legal address
  Addr -= LoadAddr;
  unsigned i = 0;
  unsigned LegalAddr = 0; // last legal address

  while (i < Addr) {
    // advance to next instruction
    LegalAddr = i;

    paramtype param = instrtable[(unsigned char)(IBuf[i])].param;

    i++;
    i += ((int) param) >> 4;  // high byte of param tells # bytes
  }

  if (i == Addr) return Addr + LoadAddr;

  return LegalAddr + LoadAddr;
}

void AddAddr(unsigned Addr) {
  if (!(Addr & 0xff00)) return; // no zero page addrs
  if ((Addr-LoadAddr) >= IBufSize) return; // out of bounds
  if (FindLabel(Addr) >= 0) return;  // already labelled

  unsigned LastValid = GetLastValidAddr(Addr);

  if (FindLabel(LastValid) < 0) {
    LabelAddrs[NumLabelAddrs] = LastValid; // add the last valid address
    LabelOffset[NumLabelAddrs++] = 0;
  }

  if (NumLabelAddrs > MAXLABELS)
    ErrStr << "Too many labels to process" << ENDE;

  if (LastValid != Addr) {
    LabelAddrs[NumLabelAddrs] = Addr;  // add the address
    LabelOffset[NumLabelAddrs++] = Addr - LastValid;
  }

  if (NumLabelAddrs > MAXLABELS)
    ErrStr << "Too many labels to process" << ENDE;
}

void LookForLabel(unsigned& CByte) {
  paramtype param = instrtable[(unsigned char)(IBuf[CByte])].param;

  CByte++;

  unsigned hex2 = IBuf[CByte];
  unsigned hex4 = * ((unsigned far *) (IBuf + CByte));

  switch (param) {
    case TYPEA: // nothing
      break;
    case TYPEB: // ($00,X)
      AddAddr((unsigned char) hex2);
      CByte++;
      break;
    case TYPEC: // $00
      AddAddr((unsigned char) hex2);
      CByte++;
      break;
    case TYPED: // #$00
      CByte++;
      break;
    case TYPEE: // ($00),Y
      AddAddr((unsigned char) hex2);
      CByte++;
      break;
    case TYPEF: // $00,X
      AddAddr((unsigned char) hex2);
      CByte++;
      break;
    case TYPEL: // $00,Y
      AddAddr((unsigned char) hex2);
      CByte++;
      break;
    case TYPEK: // $0000 for near jumps
      CByte += 1;
      hex4 = CByte + LoadAddr;
      if (hex2 & 0x80) hex4 += (hex2 & 0x7f) - 0x80;
      else hex4 += (hex2 & 0x7f);
      AddAddr((unsigned) hex4);
      break;
    case TYPEG: // $0000
      AddAddr((unsigned) hex4);
      CByte += 2;
      break;
    case TYPEH: // $0000,Y
      AddAddr((unsigned) hex4);
      CByte += 2;
      break;
    case TYPEI: // $0000,X
      AddAddr((unsigned) hex4);
      CByte += 2;
      break;
    case TYPEJ: // ($0000)
      AddAddr((unsigned) hex4);
      CByte += 2;
      break;
  }
}

void MakeLineLabels() {
  unsigned Cur = CurByte;

  int oldamtdone = 0;
  while (Cur < IBufSize) {
    int amtdone = (long) Cur * 6 / (long) IBufSize;
    if (amtdone != oldamtdone) {
      con << '.';
      oldamtdone = amtdone;
    }
    LookForLabel(Cur);
  }

  // sort the array
  for (int i = 0; i < NumLabelAddrs; i++)
    for (int j = i+1; j < NumLabelAddrs; j++) {
      if (LabelAddrs[i] >= LabelAddrs[j]) {
	unsigned t = LabelAddrs[i];
	LabelAddrs[i] = LabelAddrs[j];
	LabelAddrs[j] = t;

	// swap the offsets
	unsigned char tt= LabelOffset[i];
	LabelOffset[i] = LabelOffset[j];
	LabelOffset[j] = tt;
      }
    }
}

void WriteAddr(unsigned Addr) {
  int Label = FindLabel(Addr);
  int RealLabel;

  if (Label >= 0) {
    RealLabel = Label;
    while (LabelOffset[RealLabel]) RealLabel--;  // go back to last legal label

    cout << "l_" << RealLabel;
    if (RealLabel != Label) cout << " + " << (int) LabelOffset[Label];
  } else {
    cout << '$';
    WriteHexWord(cout,Addr);
  }
}

void WriteAddr(unsigned char Addr) {
  cout << '$';
  WriteHexByte(cout,Addr);
}

void DisassembleCurrent() {
  if (instrtable[(unsigned char)(IBuf[CurByte])].instr == QQQ) {
    cout << "dat $";
    WriteHexByte(cout,(unsigned char) IBuf[CurByte++]);
    return;
  }

  // write the instruction name:
  instrtype InstrCode = instrtable[(unsigned char)(IBuf[CurByte])].instr;

  cout << instrstr[InstrCode] << ' ';

  //now, write the parameters:
  paramtype param = instrtable[(unsigned char)(IBuf[CurByte])].param;

  CurByte++;

  unsigned hex2 = IBuf[CurByte];
  unsigned hex4 = * ((unsigned far *) (IBuf + CurByte));

  switch (param) {
    case TYPEA: // nothing
      switch (InstrCode) {
	case ASL:
	case LSR:
	case ROR:
	case ROL:
	  cout << "acc";
      }
      break;
    case TYPEB: // ($00,X)
      cout << '('; WriteAddr((unsigned char) hex2); cout << ",x)";
      CurByte++;
      break;
    case TYPEC: // $00
      WriteAddr((unsigned char) hex2);
      CurByte++;
      break;
    case TYPED: // #$00
      cout << "#$"; WriteHexByte(cout,(unsigned char) hex2);
      CurByte++;
      break;
    case TYPEE: // ($00),Y
      cout << '('; WriteAddr((unsigned char) hex2); cout << "),y";
      CurByte++;
      break;
    case TYPEF: // $00,X
      WriteAddr((unsigned char) hex2); cout << ",x";
      CurByte++;
      break;
    case TYPEL: // $00,Y
      WriteAddr((unsigned char) hex2); cout << ",y";
      CurByte++;
      break;
    case TYPEK: // $0000 for near jumps
      CurByte += 1;
      hex4 = CurByte + LoadAddr;
      if (hex2 & 0x80) hex4 += (hex2 & 0x7f) - 0x80;
      else hex4 += (hex2 & 0x7f);
      WriteAddr((unsigned) hex4);
      break;
    case TYPEG: // $0000
      WriteAddr((unsigned) hex4);
      CurByte += 2;
      break;
    case TYPEH: // $0000,Y
      WriteAddr((unsigned) hex4); cout << ",y";
      CurByte += 2;
      break;
    case TYPEI: // $0000,X
      WriteAddr((unsigned) hex4); cout << ",x";
      CurByte += 2;
      break;
    case TYPEJ: // ($0000)
      cout << '('; WriteAddr((unsigned) hex4); cout << ')';
      CurByte += 2;
      break;
  }
}


#define CHUNKSIZE 0x4000

void ReadBuf(const char *Filename,BOOL SkipLoadAddr) {
  IBuf = (char huge *) farmalloc(65536);
  if (!IBuf) ErrStr.OutOfMemory();

  ifstream(Filename,ios::binary);

  long filesize = 0;

  int infile = fasdos_open(Filename);
  if (infile < 0) ErrStr << "Unable to open input file" << ENDE;

  char huge *cur = IBuf;

  if (!SkipLoadAddr)
    fasdos_read(infile,(char *)&LoadAddr,sizeof LoadAddr);
  else LoadAddr = 0;

  // now read the data in by chunks:
  int result;
  do {
    result = fasdos_read(infile,(char *) cur,CHUNKSIZE);

    if (result < 0) ErrStr << "Problem reading input file" << ENDE;
    filesize += result;
    cur += result;
    if (filesize > 0xf000) ErrStr << "Input file is too large" << ENDE;
  } while (result == CHUNKSIZE);

  fasdos_close(infile);

  // append some zeros in case last instruction runs over:
  *cur = 0;  cur++;
  *cur = 0;  cur++;
  *cur = 0;  cur++;

  IBufSize = filesize;
  CurByte = 0;
}


void ShowHelp() {
  cout.flush();

  con << "\nCDASM - Commodore 64 Disassembler"
       << "\nVersion " << VersionID()
       << "\nWritten by ";

  if (!WRITENAME(con)) con << "?><FDDSEFD<";
  con << ", 1995"
       << "\n\nUSAGE: CDASM <SourceFile> [/A] [/B] [/L] [/N] [/R]\n\n" <<
"     CDASM disassembles a .PRG file, writing the output to\n\
the screen.  The following command-line options are available:\n\n\
/A  Shows the address of every line as a comment.\n\
/B  Shows original bytes as comments.\n\
/L  Creates line labels for addresses within the file.\n\
/N  Indicates no load address stored at beginning of file.\n\
/R  Inserts a LAYOUT section, so the output may be re-assembled with CASM.\n";

  con.flush();

  WriteFreewareMsg();

  exit(1);
}

int c_break() {
  cout.flush();
  con << "\n[CTRL-C Pressed]\n";
  return 0; // return to DOS
}

int main(int argc,char *argv[]) {
  ctrlbrk(c_break);
  int c=argc;

  BOOL LineLabels = FALSE;
  BOOL Addresses = FALSE;
  BOOL NoLoadAddress = FALSE;
  BOOL ShowOrigBytes = FALSE;
  BOOL ReAssemblable = FALSE;

  while (c > 2) {
    if (FString("/A") == argv[c-1]) Addresses = TRUE;
    else if (FString("/L") == argv[c-1]) LineLabels = TRUE;
    else if (FString("/N") == argv[c-1]) NoLoadAddress = TRUE;
    else if (FString("/B") == argv[c-1]) ShowOrigBytes = TRUE;
    else if (FString("/R") == argv[c-1]) ReAssemblable = TRUE;
    else ShowHelp();
    c--;
  }

  if (c != 2) ShowHelp();

  if (strcmp(argv[1],"/?") == 0) ShowHelp();
  if (*argv[1] == '/')
    ErrStr << "Missing filename" << ENDE;


  void (*_DisassembleCurrent)() = DisassembleCurrent;

	con << "CDASM Version " << VersionID() << ", Written by ";
	if (!WRITENAME(con)) _DisassembleCurrent = ShowHelp;
	con << ", 1995\n";
	con.flush();

	DString MainFileName(argv[1]);
  MainFileName.AllUpper();
  if (ifstream(MainFileName.c_str()).fail()) MainFileName += ".PRG";
  if (ifstream(MainFileName.c_str()).fail())
    ErrStr << "Unable to open \"" << argv[1] << "\"\n" << ENDE;

  ReadBuf(MainFileName.c_str(),NoLoadAddress);

  if (LineLabels) {
    cout.flush();  // ensure that cout finishes writing before con starts

    con << "\nCreating line labels";
    MakeLineLabels();
    con << '\n';

    con.flush();
  }

  cout << "\n\n\n;---------------------------------------------;\n"
       << ";  Disassembly for: " << MainFileName
       << "\n;---------------------------------------------;\n";
  if (!NoLoadAddress) {
    cout << ";Load Address: $";
    WriteHexWord(cout,LoadAddr);
    cout << '\n';
  }

  if (ReAssemblable) {
    cout << "\nLAYOUT\n$";
		WriteHexWord(cout,LoadAddr);
		cout << ':' <<  ((char) 9) << "Disassembly\nENDL\n";
		 cout << "\nCODE Disassembly\n";
	}

	while (CurByte < IBufSize) {
		cout << '\n';
    if (FindLabel(CurByte + LoadAddr) >= 0) {
      WriteAddr(CurByte + LoadAddr);
      cout << ':' << ((char) 9);
    } else if (Addresses) {
      if (ReAssemblable) cout << ";-";
      cout << '$';
      WriteHexWord(cout,CurByte + LoadAddr);
      if (ReAssemblable) cout << "-;  ";
      else cout << ":   "; // makes tabs work out
    } else cout << ((char) 9);
    cout << ((char) 9);

    unsigned OldCur = CurByte;
    _DisassembleCurrent();

    if (ShowOrigBytes) {
			cout << ' ' << ((char) 9);
      if (CurByte-OldCur == 1) cout << ((char) 9); // extra tab for short ones
      cout << ((char) 9) << ";  ";
      while (OldCur < CurByte) {
	WriteHexByte(cout,(unsigned char)IBuf[OldCur++]);
	cout << ' ';
      }
    }
  }
  if (ReAssemblable) cout << "\nENDC\n";

  cout << "\n\n;---------------------------------------------;\n";
  return 0;
}


