/*
  CDUMP.CPP - Written by Peter Gonzalez [pbgonz@mail.wm.edu], 1995

  This is just one big mess of output formatting.

    CASM, CDASM, CLINK, and CDUMP are freeware; they may be freely
  distributed, provided that no fee is charged and I am given due
  credit for my work.  This source code is available directly from me upon
  request; please do not distribute the source code or any compiled
  modifications without my consent.  Support freeware!
*/

#include <fstream.h>
#include <iomanip.h>
#include <dos.h>

#include "diskacc.h"
#include "version.h"
#include "numbers.h"

char huge *IBuf;
unsigned IBufSize;
unsigned CurByte;  // offset into IBuf
unsigned LoadAddr;

void WriteBinaryByte(unsigned char byte,char OneChar) {
  for (int i=1;i<=8;i++) {
    cout << ((byte & 0x80) ? OneChar : '.');
    byte <<= 1;
  }
}

#define CHUNKSIZE 0x1000

void ReadBuf(const char *Filename,BOOL SkipLoadAddr) {
  IBuf = (char huge *) farmalloc(65536);
  if (!IBuf) ErrStr.OutOfMemory();

  ifstream(Filename,ios::binary);

  long filesize = 0;

  int infile = fasdos_open(Filename);
  if (infile < 0) ErrStr << "Unable to open input file" << ENDE;

  char huge *cur = IBuf;

  if (!SkipLoadAddr)
    fasdos_read(infile,(char *)&LoadAddr,sizeof LoadAddr);
  else LoadAddr = 0;

  // now read the data in by chunks:
  int result;
  do {
    result = fasdos_read(infile,(char *) cur,CHUNKSIZE);

    if (result < 0) ErrStr << "Problem reading input file" << ENDE;
    filesize += result;
    cur += result;
    if (filesize > 0xf000) ErrStr << "Input file is too large" << ENDE;
  } while (result == CHUNKSIZE);

  fasdos_close(infile);

  // append some zeros in case last instruction runs over:
  *cur = 0;  cur++;
  *cur = 0;  cur++;
  *cur = 0;  cur++;

  IBufSize = filesize;
  CurByte = 0;
}


void ShowHelp() {
  con << "\nCDUMP - Commodore 64 Hex Dump Utility"
       << "\nVersion " << VersionID() << ", written by ";

  if (!WRITENAME(con)) con << "?><FDDSEFD<";
  con << ", 1995"

       << "\n\nUSAGE: CDUMP <SourceFile> [/A] [/B#] [/D] [/N] [/X[,YY]]\n\n" <<
"     CDUMP displays dumps the contents of a file to the screen.  By \n\
default, the output is shown in rows of 8 hexadecimal numbers.  The\n\
following command-line options are available:\n\n\
/A     Shows addresses for each line.\n\
/B#    Shows the numbers as binary bitmaps.  # can be \"*\", \"x\", or \"X\"\n\
/D     Shows the numbers in decimal.\n\
/N     Indicates no load address stored in file.\n\
/X,YY  X tells number of bytes per line, YY tells number of blocks.\n";
  con.flush();

  WriteFreewareMsg();

  exit(1);
}

int c_break() {
  cout.flush();
  con << "\n[CTRL-C Pressed]";
  return 0; // return to DOS
}

int main(int argc,char *argv[]) {
  ctrlbrk(c_break);

  int c=argc;

  BOOL Addresses = FALSE;
  enum { DECIMAL, BINARY, HEX } Base = HEX;
  BOOL NoLoadAddress = FALSE;
  int WrapSize = 8;
  int BlockSize = 0;
  char OneChar;

  while (c > 2) {
    // really ugly routine for parsing /X,YY

    char *ch = argv[c-1];
    if (*ch != '/') goto nope;
    ch++;
    if ((*ch < '1') || (*ch > '9')) goto nope;
    WrapSize = *ch - '0';
    ch++;
    if (!*ch) { c--; continue; }
    if (*ch != ',') goto nope;
    ch++;
    if ((*ch < '0') || (*ch > '9')) goto nope;
    BlockSize = *ch - '0';
    ch++;
    if (!*ch) { c--; continue; }
    if ((*ch < '0') || (*ch > '9')) goto nope;
    BlockSize = (BlockSize*10) + (*ch-'0');
    ch++;
    if (!*ch) { c--; continue; }

    nope:

    ch = argv[c-1];
    if (*ch != '/') goto nope2;
    ch++;
    if ((*ch != 'B') && (*ch != 'b')) goto nope2;
    ch++;
    if (*(ch+1)) goto nope2;
    switch (*ch) {
      case '*':
      case 'x':
      case 'X':
	Base = BINARY;
	OneChar = *ch;
	c--;
	continue;
    }

    nope2:;


    if ((argv[c-1][0] == '/') && (!argv[c-1][2])) {
      switch (argv[c-1][1]) {
	case 'D':
	case 'd':
	  Base = DECIMAL; break;
	case 'N':
	case 'n':
	  NoLoadAddress = TRUE; break;
	case 'A':
	case 'a':
	  Addresses = TRUE; break;
	default:
	  ShowHelp();
      }
    } else ShowHelp();
    c--;
  }

  if (c != 2) ShowHelp();

  if (strcmp(argv[1],"/?") == 0) ShowHelp();
  if (*argv[1] == '/')
    ErrStr << "Missing filename" << ENDE;

  con << "CDUMP Version " << VersionID() << ", Written by ";
  if (WRITENAME(con)); // do nothing
  con << ", 1995\n";

  con.flush();

  DString MainFileName(argv[1]);
  MainFileName.AllUpper();
  if (ifstream(MainFileName.c_str()).fail()) MainFileName += ".PRG";
  if (ifstream(MainFileName.c_str()).fail())
    ErrStr << "Unable to open \"" << argv[1] << "\"\n" << ENDE;

  ReadBuf(MainFileName.c_str(),NoLoadAddress);

  if (NoLoadAddress) LoadAddr = 0;
  else if (!Addresses) {
    cout << "\nLoad Address: $";
    WriteHexWord(cout,LoadAddr);
    cout << '\n';
  }


  int wrap = 0;
  int blockline = 0;

  cout.fill('0');

  while (IBufSize--) {
    if (!wrap) {
      cout << "\n";

      if (BlockSize) {
	if (!blockline) cout << "\n";
	++blockline %= BlockSize;
      }

      if (Addresses) {
	cout << "  $";
	WriteHexWord(cout,CurByte + LoadAddr);
	cout << ":";
      }
      cout << ((char) 9);
      if (Base == BINARY) cout << '[';
    }
    ++wrap %= WrapSize;

    switch (Base) {
      case DECIMAL:
	cout << setw(3) << (unsigned) (unsigned char) IBuf[CurByte++];
	break;
      case HEX:
	cout << '$';
	WriteHexByte(cout,(unsigned char) IBuf[CurByte++]);
	break;
      case BINARY:
	WriteBinaryByte(IBuf[CurByte++],OneChar);
	break;
    }

    if (!wrap) {
      if (Base == BINARY) cout << ']';
    } else if (IBufSize) cout << ' ';
  }

  if ((wrap) && (Base == BINARY)) cout << ']';
  cout << "\n";
  cout.flush();
  con << "\n";
  return 0;
}


