/*
  CLINK.CPP - Written by Peter Gonzalez [pbgonz@mail.wm.edu], 1995

    CASM, CDASM, CLINK, and CDUMP are freeware; they may be freely
  distributed, provided that no fee is charged and I am given due
  credit for my work.  This source code is available directly from me upon
  request; please do not distribute the source code or any compiled
  modifications without my consent.  Support freeware!
*/

#include <iostream.h>
#include <alloc.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <dir.h>
#include <dos.h>

#include "tokenize.h"
#include "tables.h"
#include "diskacc.h"
#include "numbers.h"
#include "version.h"

char huge *BigBuf;
unsigned BigBufSize;

DString LSTFileName;

void WriteFile(DString& Filename,char *buf,unsigned size,unsigned loadaddr) {
  unsigned OutFile = fasdos_create(Filename.c_str(),0);
  if (OutFile == 0xffff) {
    ErrStr << "Unable to open output file \"" << Filename << '\"' << ENDE;
  }

  fasdos_write(OutFile,(char *) &loadaddr,2);

  while (size > 0) {
    unsigned blocksize;
    if (size < 0x4000) blocksize = size;
    else blocksize = 0x4000;
    if (fasdos_write(OutFile,(char far *)buf,blocksize) == 0xffff) {
      fasdos_close(OutFile);
      ErrStr << "Error writing to output file \""
	<< Filename << '\"' << ENDE;
    }
    size -= blocksize;
    buf += blocksize;
  }

  fasdos_close(OutFile);
}

void ReadFile(DString& Filename,char *buf,unsigned& size,unsigned& loadaddr) {

  unsigned InFile = fasdos_open(Filename.c_str());
  if (InFile == 0xffff) {
    ErrStr << "Unable to open input file \"" << Filename << '\"' << ENDE;
  }

  if (fasdos_read(InFile,(char *) &loadaddr,2) != 2) {
      fasdos_close(InFile);
      ErrStr << "Error writing to input file \""
	<< Filename << '\"' << ENDE;
  }

  size = 0;
  unsigned amtread;

  do {
    amtread = fasdos_read(InFile,(char far *)buf,0x4000);

    if (amtread == 0xffff) {
      fasdos_close(InFile);
      ErrStr << "Error reading input file \"" << Filename << '\"' << ENDE;
    }

    if ((size + (long)amtread) > 0xf000) {
      fasdos_close(InFile);
      ErrStr << "File \"" << Filename << "\" is too big to read" << ENDE;
    }

    size += amtread;
    buf += 0x4000;
  } while (amtread == 0x4000);

  fasdos_close(InFile);
}

unsigned long GetPackedFileTimeDate(const char *FileName) {
  struct ffblk ffblk;
  if (findfirst(FileName,&ffblk,0)) return 0;
  return (((unsigned long)ffblk.ff_fdate) << 16) + ffblk.ff_ftime;
}

void Assemble(DString& PRGFileName) {
  DString ASMFileName = PRGFileName;
  ASMFileName.RemoveFileExtension();
  ASMFileName += ".ASM";

  // now compare file dates:
  unsigned long ASMdate = GetPackedFileTimeDate(ASMFileName.c_str());
  unsigned long PRGdate = GetPackedFileTimeDate(PRGFileName.c_str());

  // don't reassemble if dates don't match:
  if ((PRGdate) && (ASMdate <= PRGdate)) return;

  cout.flush();
  int ret = spawnlp(P_WAIT,"CASM.EXE","CASM.EXE",ASMFileName.c_str(),
    "/Q",NULL);
  if (ret == -1) {
    ErrStr << "Error executing CASM.EXE: ";
    if (!errno) return;
    switch (errno) {
      case ENOENT: ErrStr << "File not found" << ENDE;
      case ENOMEM: ErrStr << "Not enough conventional memory" << ENDE;
      default: ErrStr << "DOS Error #" << errno << ENDE;
    }
  }
  if (ret) {
    cout << "\n(CASM error; CLINK is terminating.)\n";
    exit(1);
  }
}

void LinkFile(BOOL AssembleFirst) {
  cout << "\nNow linking " << LSTFileName << "...";
  if (AssembleFirst) cout << '\n';

  TASMFileTokenizer ftok(LSTFileName);
  TToken ModuleName;
  TToken PrevModuleName;

  unsigned LoadAddr = ReadWord(ftok);
  unsigned ModuleLoadAddr;
  unsigned ModuleAddr;

  do {
    ftok.SkipWhite();
    if (ftok.Eof()) break;
    ModuleName = FString(ftok.ReadToDelim(" \n\t"));
    if (ModuleName == "") break;
    ModuleName.Text.AllUpper();

    ModuleAddr = ReadWord(ftok);

    if (ModuleAddr < LoadAddr) {
      ModuleName.WriteFileLine(ErrStr);
      ErrStr << ModuleName
	<< "'s address is less than the load address" << ENDE;
    }

    if (ModuleAddr < (BigBufSize + LoadAddr)) {
      ModuleName.WriteFileLine(ErrStr);
      ErrStr << PrevModuleName
	<< " runs past " << ModuleName << "'s address" << ENDE;
    }

    if (AssembleFirst) Assemble(DString(ModuleName.Text));

    // now read ModuleName into BigBuf at ModuleAddr:
    if (BigBufSize < (ModuleAddr - LoadAddr))
      BigBufSize = ModuleAddr - LoadAddr;

    unsigned amtread;
    // if BigBuf overflows, it will wrap around because it's a far pointer,
    // so there won't be a problem:
    ReadFile(DString(ModuleName.Text),
      ((char far *)BigBuf) + BigBufSize,amtread,ModuleLoadAddr);
    if ((amtread + (long)BigBufSize) > 0xffffL) ErrStr << ModuleName
      << " spills over the address $FFFF" << ENDE;

    if (ModuleLoadAddr != ModuleAddr) {
      cout << "\nWarning: " << ModuleName
	<< "'s load address does not match the address in " << LSTFileName;
    }
    BigBufSize += amtread;
    PrevModuleName = ModuleName;
  } while(1);

  LSTFileName.RemoveFileExtension();
  LSTFileName += ".PRG";

  cout << "\nWriting " << LSTFileName << "...";
  WriteFile(LSTFileName,(char far*)BigBuf,BigBufSize,LoadAddr);
  cout << "\n";
}

void UnLinkFile() {
  cout << "\nNow unlinking " << LSTFileName << "...";

  DString PRGFileName = LSTFileName;
  PRGFileName.RemoveFileExtension();
  PRGFileName += ".PRG";

  unsigned FileLoadAddr;
  ReadFile(PRGFileName,(char far *)BigBuf,BigBufSize,FileLoadAddr);

  TASMFileTokenizer ftok(LSTFileName);
  TToken ModuleName;

  unsigned LoadAddr = ReadWord(ftok);
  unsigned ModuleAddr;

  TToken PrevModuleName;  // name and start addr of previous module
  unsigned PrevAddr;

  if (FileLoadAddr != LoadAddr) {
    cout << "\nWarning: " << PRGFileName
      << "'s load address does not match the address in " << LSTFileName;
  }

  PrevModuleName.Text = "";
  PrevAddr = LoadAddr;

  do {
    ftok.SkipWhite();
    if (ftok.Eof()) break;
    ModuleName = FString(ftok.ReadToDelim(" \n\t"));
    if (ModuleName == "") break;
    ModuleName.Text.AllUpper();

    ModuleAddr = ReadWord(ftok);

    if (ModuleAddr < LoadAddr) {
      ModuleName.WriteFileLine(ErrStr);
      ErrStr << ModuleName
	<< "'s address is less than the load address" << ENDE;
    }

    if (ModuleAddr > (BigBufSize + LoadAddr)) {
      ModuleName.WriteFileLine(ErrStr);
      ErrStr << PrevModuleName
	<< "'s address is past end of " << PRGFileName << ENDE;
    }

    if (PrevModuleName != "") { // this is not the first module
      // write out the previous module, up to this module's address:
      if (ModuleAddr < PrevAddr) {
	ModuleName.WriteFileLine(ErrStr);
	ErrStr << PrevModuleName
	  << " runs past " << ModuleName << "'s start address" << ENDE;
      }

      unsigned Amt2Write = ModuleAddr - PrevAddr;
      // now write out the module's data:
      WriteFile(DString(PrevModuleName.Text),
        (char far*)BigBuf + (PrevAddr - LoadAddr),
	Amt2Write,PrevAddr);
    }

    PrevModuleName = ModuleName;
    PrevAddr = ModuleAddr;
  } while(1);

  // last, write out PrevModuleName up to the end of the file:
  if (PrevModuleName != "") {
    WriteFile(DString(PrevModuleName.Text),
      (char far*)BigBuf + (PrevAddr - LoadAddr),
      BigBufSize - (PrevAddr - LoadAddr),PrevAddr);
  }
  cout << "\nFile unlinked.\n";
}

void ShowHelp() {
  cout << "\nCLINK - Multi-Module Project Manager"
       << "\nVersion " << VersionID()
       << ", written by ";

  if (!WRITENAME()) cout << "?><FDDSEFD<";
  cout << ", 1995"
       << "\nUSAGE: CLINK <ListFile> [/U|/C]\n\n" << ((char) 9) <<
"CLINK pastes together a set of files according to the specifications\n\
given in the .LST list file, and stores the result in a .PRG file with\n\
the same name as the .LST file.  The \"/U\" parameter reverses the process,\n\
generating small files from the .PRG file.  The \"/C\" parameter runs CASM\n\
on the .ASM files whose dates/time are earlier than that of the .PRG\n\
file.\n\n\
Here is an example .LST file:\n\n"
       << ((char) 9) << ((char) 9) << "$1000  ;Load address\n"
       << "FILE1.PRG" << ((char) 9) << "$1000  ;Start addresses\n"
       << "FILE2.PRG" << ((char) 9) << "$1500\n";

  WriteFreewareMsg();
  exit(1);
}

int c_break() {
  cout << "\n[CTRL-C Pressed]";
  return 0; // return to DOS
}

int main(int argc,char *argv[]) {
  ctrlbrk(c_break);
  int c=argc;

  BOOL UnLink = FALSE;
  BOOL AssembleFirst = FALSE;
  while (c > 2) {
    if (FString("/U") == argv[c-1]) UnLink = TRUE;
    else if (FString("/C") == argv[c-1]) AssembleFirst = TRUE;
    else ShowHelp();
    c--;
  }

  if (AssembleFirst && UnLink) {
    cout << "\nInvalid combination of parameters";
    ShowHelp();
  }

  if (c != 2) ShowHelp();

  strupr(argv[1]);
  LSTFileName = argv[1];

  if (strcmp(argv[1],"/?") == 0) ShowHelp();
  if (*argv[1] == '/')
    ErrStr << "Missing filename" << ENDE;


  cout << "CLINK Version " << VersionID() << ", Written by ";
  if (!WRITENAME()) BigBuf += 5;
  cout << ", 1995\n";

  if (ifstream(argv[1]).fail()) LSTFileName += ".LST";
  if (ifstream(LSTFileName.c_str()).fail())
    ErrStr << "Unable to open \"" << argv[1] << "\"\n" << ENDE;

  BigBuf = (char huge *) farmalloc(0x20000); // add some safety space
  memset(BigBuf,0,0xffff);
  if (!BigBuf) ErrStr.OutOfMemory();
  BigBufSize = 0;

  if (UnLink) UnLinkFile();
  else LinkFile(AssembleFirst);

  farfree(BigBuf);

  return 0;
}

