/* 
  DISKACC.CPP - Written by Peter Gonzalez [pbgonz@mail.wm.edu], 1995 
 
    CASM, CDASM, CLINK, and CDUMP are freeware; they may be freely 
  distributed, provided that no fee is charged and I am given due 
  credit for my work.  This source code is available directly from me upon 
  request; please do not distribute the source code or any compiled  
  modifications without my consent.  Support freeware!  
*/ 
 
#include <dir.h>
#include <dos.h>

int existfile(const char *filename,int attrib) {
/*
  Returns nonzero if a file matching ~filename~ can be found.
  Wildcards are allowed.

  attrib, which identifies which types of files to look for, should be
  some combination of these DOS.H constants:

    Constant   Description
   
    FA_RDONLY  Read-only attribute
    FA_HIDDEN  Hidden file
    FA_SYSTEM  System file
    FA_LABEL   Volume label
    FA_DIREC   Directory
    FA_ARCH    Archive
*/
  struct ffblk ffblk;
  return !findfirst(filename,&ffblk,attrib);
}

int deletefile(const char *filename) {
/*
  deletefile() attempts to delete the file ~filename~, and returns
  zero if successful.  If an error occurs, a DOS error code is returned.
*/
  unsigned file_seg = FP_SEG(filename);
  unsigned file_off = FP_OFF(filename);
  asm {
	push	ds
	mov	ah,0x41
	mov	dx,file_off
	mov	ds,file_seg
	int	0x21
	jc	error
	pop	ds
  }
  return 0;
  int err;
  error:;
  asm {
	mov	err,ax
	pop	ds
  }
  return err;
}
/*  removed because it's not used and there is no need to link in the floating
    point library
    
long diskfreespace(char drive) {
/*
  This function accurately returns the number of bytes of free space on a
  hard drive up to 1.9 GB.  Any larger values are returned as 2147483647.
  diskfreespace() returns -1 if ~drive~ is invalid or some other error
  occurs.

  ~drive~ should be an ASCII letter.
* /
  unsigned availableclusters;
  unsigned sectorspercluster;
  unsigned bytespersector;

  // convert drive from a letter to a code:
  if (drive >= 'a') drive -= 32;  // make uppercase
  if ((drive < 'A') || (drive > 'Z')) return -1;
  drive -= '@';

  // ~drive~ is now equal to 1 for A, 2 for B, 3 for C, etc.

  asm {
	mov	ah,0x36
	mov	dl,drive
	int	0x21
	mov	availableclusters,bx
	mov	bytespersector,cx
	mov	sectorspercluster,ax
  }
  if (sectorspercluster == 0xffff) return -1;  // an error occurred

  long double checkproduct = (long double) availableclusters
    * (long double) sectorspercluster
    * (long double) bytespersector;

  // this prevents incredibly huge drives from exceeding MAXLONG
  if (checkproduct > 2147483647L) return 2147483647L;

  return (long) availableclusters
    * (long) sectorspercluster
    * (long) bytespersector;
}
*/
unsigned fasdos_write(int handle, void *buf, unsigned len) {
/*
  Writes data to binary files like Borland's write(), only a lot faster.
  Returns number of bytes written or -1 for error.
*/
  unsigned buf_seg = FP_SEG(buf);
  unsigned buf_off = FP_OFF(buf);
  int retval;

  asm {
	push	ds
	mov	ah,0x40
	mov	bx,handle
	mov	cx,len
	mov	dx,buf_off
	mov	ds,buf_seg
	int 	0x21
	jc	error
	mov	retval,ax
	pop	ds
  }
  return retval;
  error:;
  asm {
	pop	ds
  }
  return -1;
}

unsigned fasdos_read(int handle, void *buf, unsigned len) {
/*
  Reads data from binary files like Borland's read(), only a lot faster.
  Returns number of bytes read or -1 for error.
*/
  unsigned buf_seg = FP_SEG(buf);
  unsigned buf_off = FP_OFF(buf);
  int retval;

  asm {
	push	ds
	mov	ah,0x3f
	mov	bx,handle
	mov	cx,len
	mov	dx,buf_off
	mov	ds,buf_seg
	int 	0x21
	jc	error
	mov	retval,ax
	pop	ds
  }
  return retval;
  error:;
  asm {
	pop	ds
  }
  return -1;
}

unsigned fasdos_movefp(int handle,unsigned long offsetfromstart) {
/*
  Returns 0 if successful, or -1 for error.
*/
  unsigned start_hi = offsetfromstart >> 16;
  unsigned start_lo = offsetfromstart & 0xffff;

  asm {
	mov	ax,0x4200
	mov	bx,handle
	mov	cx,start_hi
	mov	dx,start_lo
	int	0x21
	jc	error
	cmp	start_hi,dx
	jne	error
	cmp	start_lo,ax
	jne	error
  }
  return 0;
  error:;
  return -1;
}

unsigned fasdos_create(const char *filename, int mode) {
  unsigned file_seg = FP_SEG(filename);
  unsigned file_off = FP_OFF(filename);

  int newhandle;

  asm {
	push	ds
	mov	ah,0x3c
	mov	cx,mode
	mov	dx,file_off
	mov	ds,file_seg
	int	0x21
	jc	error
	mov	newhandle,ax
	pop	ds
  }
  return newhandle;
  error:;
  asm {
	pop	ds
  }
  return -1;
}

unsigned fasdos_open(const char *filename) {
  unsigned file_seg = FP_SEG(filename);
  unsigned file_off = FP_OFF(filename);

  int newhandle;

  asm {
	push	ds
	mov	ax,0x3d00   // 00 = normal handle access code
	mov	dx,file_off
	mov	ds,file_seg
	int	0x21
	jc	error
	mov	newhandle,ax
	pop	ds
  }
  return newhandle;
  error:;
  asm {
	pop	ds
  }
  return -1;
}

unsigned fasdos_close(int handle) {
  asm {
	mov	ah,0x3e
	mov	bx,handle
	int	0x21
	jc	error
  }
  return 0;
  error:;
  return -1;
}
