/* 
  FIXUP.CPP - Written by Peter Gonzalez [pbgonz@mail.wm.edu], 1995

    The functions defined here are called after all blocks have been read.
  They complete the calculations which could not be done during the reading
  process.
 
    CASM, CDASM, CLINK, and CDUMP are freeware; they may be freely 
  distributed, provided that no fee is charged and I am given due 
  credit for my work.  This source code is available directly from me upon 
  request; please do not distribute the source code or any compiled  
  modifications without my consent.  Support freeware!  
*/ 
 
#include "block.h"
#include "tokenize.h"
#include "casm.h"

//--------------------------------------------------------------------------
void __ResolveLabel(TBlock::TLabelRef *Ref,TBlock *RefBlock) {
//--------------------------------------------------------------------------
// isolate all the variable creations here, away from the stack frame

// RefBlock is the block where Ref is referenced
// DefBlock is the block where the label referenced is defined

  TBlock *DefBlock = Ref->DefBlock;
  TBlock::TLabelDef *Def = (TBlock::TLabelDef *)
    DefBlock->LabelDefs.Find(Ref->Name.c_str());

  if (Def) {
    // found it
    unsigned DestIndex = Ref->OffsetInBlock + RefBlock->OffsetInDestFile;

    unsigned AddrToStore = Def->OffsetInBlock + DefBlock->OffsetInDestFile
      + GP.ProgStart + Ref->AddAmount;

    if (Ref->RefType == TBlock::TLabelRef::FAR) {
      *((unsigned *) (GP.OBuf + DestIndex)) = AddrToStore;
    } else if (Ref->RefType == TBlock::TLabelRef::HI) {
      *((unsigned char *) (GP.OBuf + DestIndex)) = AddrToStore >> 8;
    } else if (Ref->RefType == TBlock::TLabelRef::LO) {
      *((unsigned char *) (GP.OBuf + DestIndex)) = AddrToStore & 0xff;
    } else {
      // it's a NEAR branch
      // start from addr after
      unsigned StartAddr = DestIndex + GP.ProgStart + 1;

      unsigned byte;

      if (AddrToStore < StartAddr) {
	byte = AddrToStore + 0x80 - StartAddr;
	if (byte & 0xFF80) {
	  TToken tok; tok.Filename = Ref->Filename; tok.LineNum = Ref->LineNum;
	  tok.WriteFileLine(ErrStr);
	  ErrStr << "NEAR jump out of range" << ENDE;
	}
	byte |= 0x80;
      } else {
	byte = AddrToStore - StartAddr;
	if (byte & 0xFF80) {
	  TToken tok; tok.Filename = Ref->Filename; tok.LineNum = Ref->LineNum;
	  tok.WriteFileLine(ErrStr);
	  ErrStr << "NEAR jump out of range" << ENDE;
	}
      }
      GP.OBuf[DestIndex] = byte;
    }
    return;
  }

  // Maybe it's a block name:
  TBlock *LabelledBlock = (TBlock *) BlockList.Find(Ref->Name.c_str());

  if (LabelledBlock) {
    unsigned DestIndex = Ref->OffsetInBlock + RefBlock->OffsetInDestFile;

    unsigned AddrToStore = LabelledBlock->OffsetInDestFile
      + GP.ProgStart + Ref->AddAmount;

    if (Ref->RefType == TBlock::TLabelRef::FAR) {
      *((unsigned *) (GP.OBuf + DestIndex)) = AddrToStore;
    } else if (Ref->RefType == TBlock::TLabelRef::HI) {
      *((unsigned char *) (GP.OBuf + DestIndex)) = AddrToStore >> 8;
    } else if (Ref->RefType == TBlock::TLabelRef::LO) {
      *((unsigned char *) (GP.OBuf + DestIndex)) = AddrToStore & 0xff;
    } else {
      // it's a NEAR branch
      // start from addr after
      unsigned StartAddr = DestIndex + GP.ProgStart + 1;

      unsigned byte;

      if (AddrToStore < StartAddr) {
	byte = AddrToStore + 0x80 - StartAddr;
	if (byte & 0xFF80) {
	  TToken tok; tok.Filename = Ref->Filename; tok.LineNum = Ref->LineNum;
	  tok.WriteFileLine(ErrStr);
	  ErrStr << "NEAR jump out of range" << ENDE;
	}
	byte |= 0x80;
      } else {
	byte = AddrToStore - StartAddr;
	if (byte & 0xFF80) {
	  TToken tok; tok.Filename = Ref->Filename; tok.LineNum = Ref->LineNum;
	  tok.WriteFileLine(ErrStr);
	  ErrStr << "NEAR jump out of range" << ENDE;
	}
      }
      GP.OBuf[DestIndex] = byte;
    }
    return;
  }

  TToken tok; tok.Filename = Ref->Filename; tok.LineNum = Ref->LineNum;
  tok.WriteFileLine(ErrStr);

  if (Ref->Name[0] != '~')
    ErrStr << "The label \"" << Ref->Name << "\" is not defined" << ENDE;
  else ErrStr << "Reference to undefined temp label" << ENDE;
}

//--------------------------------------------------------------------------
void _ResolveLabel(TBlock::TLabelRef *Ref,TBlock *RefBlock) {
//--------------------------------------------------------------------------
  if (!Ref) return;

  __ResolveLabel(Ref,RefBlock);

  for (int i=0; i<BTREENUMBRANCHES; i++)
    _ResolveLabel((TBlock::TLabelRef *) (Ref->Branch[i]),RefBlock);
}

//--------------------------------------------------------------------------
void _ResolveLabels(THashObject *CurBlock) {
//--------------------------------------------------------------------------
  if (!CurBlock) return;

  _ResolveLabel((TBlock::TLabelRef *)
    ((TBlock*)CurBlock)->LabelRefs.GetRoot(),(TBlock*)CurBlock);

  for (int i=0; i<BTREENUMBRANCHES; i++)
    _ResolveLabels(((TBlock*)CurBlock)->Branch[i]);
}

//--------------------------------------------------------------------------
void ResolveLabels() {
//--------------------------------------------------------------------------
  _ResolveLabels(BlockList.GetRoot());
}

//--------------------------------------------------------------------------
void _FixNearJumps(THashObject *_NJ) {
//--------------------------------------------------------------------------
  if (!_NJ) return;

  TNearJump *NJ = (TNearJump *)_NJ;

  unsigned StartOffset = NJ->StartOffset
    + NJ->ParentBlock->OffsetInDestFile;
  unsigned StartAddr = StartOffset + GP.ProgStart + 1; // start from addr after

  unsigned byte;

  if (NJ->EndAddr < StartAddr) {
    byte = NJ->EndAddr + 0x80 - StartAddr;
    if (byte & 0xFF80) {
      TToken tok; tok.Filename = NJ->Filename; tok.LineNum = NJ->LineNum;
      tok.WriteFileLine(ErrStr);
      ErrStr << "NEAR jump out of range" << ENDE;
    }
    byte |= 0x80;
  } else {
    byte = NJ->EndAddr - StartAddr;
    if (byte & 0xFF80) {
      TToken tok; tok.Filename = NJ->Filename; tok.LineNum = NJ->LineNum;
      tok.WriteFileLine(ErrStr);
      ErrStr << "NEAR jump out of range" << ENDE;
    }
  }

  GP.OBuf[StartOffset] = (char) byte;

  for (int i=0; i<BTREENUMBRANCHES; i++)
    _FixNearJumps(NJ->Branch[i]);
}

//--------------------------------------------------------------------------
void FixNearJumps() {
//--------------------------------------------------------------------------
  _FixNearJumps(NearJumpList.GetRoot());
}

