/* 
  HASH.CPP - Written by Peter Gonzalez [pbgonz@mail.wm.edu], 1995 
 
    CASM, CDASM, CLINK, and CDUMP are freeware; they may be freely 
  distributed, provided that no fee is charged and I am given due
  credit for my work.  This source code is available directly from me upon
  request; please do not distribute the source code or any compiled
  modifications without my consent.  Support freeware!
*/

#include "hash.h"

#include <stdlib.h>
#include <ctype.h>

//--------------------------------------------------------------------------
unsigned THash::MakeHashValue(const char *cur) {
//--------------------------------------------------------------------------
/*
  Generates a 16-bit number from a given string, such that any two strings
  differing by one character will have different hash values.
*/
  // generate a hash value:
  unsigned long hash = 0;

  if (*cur) { hash = tolower(*cur); cur++; }
  if (*cur) { hash |= ((unsigned)tolower(*cur)) << 8; cur++; }
  while (*cur) {
    // mix it up a bit:  (roll 3 bits to left)
    hash <<= 3;
    hash ^= hash >> 19;
    hash &= 0xffff;

    // add in cur bit of data:
    hash ^= tolower(*cur);
    cur++;
  }
  return (unsigned) hash;
}

//--------------------------------------------------------------------------
unsigned THash::GetDepth(const THashObject *loc) {
//--------------------------------------------------------------------------
  if (!loc) return 0;

  unsigned max = 0;
  for (int i=0; i<BTREENUMBRANCHES; i++) {
    unsigned temp = GetDepth(loc->Branch[i]);
    if (temp > max) max = temp;
  }

  return max + 1;
}


//--------------------------------------------------------------------------
void THash::Insert(THashObject *NewObj) {
//--------------------------------------------------------------------------
  // set up its data:
  NewObj->HashValue = MakeHashValue(NewObj->Name.c_str());

  for (int i=0; i < BTREENUMBRANCHES; i++) NewObj->Branch[i] = 0;

  // now insert NewObj
  THashObject **curp = &root;

  int index;
  int depth = 0;
  srand(NewObj->HashValue);

  while (*curp) {
    depth++;
    index = (random(0x253) ^ ((*curp)->HashValue)) % BTREENUMBRANCHES;

    curp = &( (*curp)->Branch[index] );
  }

  if (depth > MAXBTREEDEPTH)
    ErrStr << "B-tree is over-extended" << PROGBUG;

  *curp = NewObj;
}

//--------------------------------------------------------------------------
THashObject *THash::Find(const char *key) {
//--------------------------------------------------------------------------
  THashObject **curp = &root;

  unsigned keyhash = MakeHashValue(key);

  int index;
  srand(keyhash);

  while (*curp) {
    if ((*curp)->HashValue == keyhash) {
      // this might be it:
      if (fastcmp(key,(*curp)->Name.c_str())) {
	return *curp;
      }
    }

    index = (random(0x253) ^ ((*curp)->HashValue)) % BTREENUMBRANCHES;

    curp = &( (*curp)->Branch[index] );
  }

  return 0;
}

//--------------------------------------------------------------------------
void THash::EmptyTree(THashObject *loc) {
//--------------------------------------------------------------------------
  if (!loc) return;
  for (int i=0; i < BTREENUMBRANCHES; i++)
    EmptyTree(loc->Branch[i]);
  delete loc;
}
