/*
  LAYOUT.CPP - Written by Peter Gonzalez [pbgonz@mail.wm.edu], 1995

    CASM, CDASM, CLINK, and CDUMP are freeware; they may be freely
  distributed, provided that no fee is charged and I am given due
  credit for my work.  This source code is available directly from me upon
  request; please do not distribute the source code or any compiled
  modifications without my consent.  Support freeware!
*/

#include "layout.h"
#include "misc.h"
#include "block.h"
#include "numbers.h"
#include "casm.h"


TBlockLayoutList *BlockLayoutList = 0;


//--------------------------------------------------------------------------
void PreScanLayoutTable(TASMFileTokenizer& ftok) {
//--------------------------------------------------------------------------
/*
  Find out which blocks need to be assembled.
*/

  TToken cur;

  cur = ftok.GetNextToken();
  if (cur != "LAYOUT")
    ErrStr << "\"LAYOUT\" expected, not \"" << cur << '"' << ENDE;

  do {
    cur = ftok.GetNextToken();
    if (cur == '$') continue;
    if (cur == "ENDL") return;
    if (ftok.Peek() == ':') {
      // skip this; it's only an address
      ftok.GetNextToken();
      continue;
    }

    if (!IsID(cur.Text)) {
      cur.WriteFileLine(ErrStr);
      ErrStr << "Invalid characters in identifier" << ENDE;
    }

    if (IsNameInUse(cur.Text)) {
      cur.WriteFileLine(ErrStr);
      ErrStr << "The identifier \"" << cur.Text << "\" is already in use"
	<< ENDE;
    }
    BlockList.Insert(new TBlock(cur.Text.c_str()));
  } while(cur != "");
}

//--------------------------------------------------------------------------
void _CheckDefinitions(THashObject *cur) {
//--------------------------------------------------------------------------
  if (!cur) return;
  if (!((TBlock *)cur)->DefinedYet)
    ErrStr << "The block \"" << ((TBlock *)cur)->Name
      << "\" listed in the LAYOUT section\nwas not defined" << ENDE;
  for (int i=0; i<BTREENUMBRANCHES; i++)
    _CheckDefinitions(cur->Branch[i]);
}

//--------------------------------------------------------------------------
void CheckDefinitions() {
//--------------------------------------------------------------------------
  THashObject *cur = BlockList.GetRoot();
  _CheckDefinitions(cur);
}

//--------------------------------------------------------------------------
void ReScanLayoutTable(const char far *Filename) {
//--------------------------------------------------------------------------
// Copy the blocks into the buffer according to the LAYOUT table listing.
// Also outputs disassembly if flagged.

  TASMFileTokenizer ftok(Filename);

  CheckDefinitions();

  TToken cur;

  TBlock *CurBlock = 0;
  TBlock *PrevBlock;

  TBlockLayoutList *CurBLL;
  TBlockLayoutList **PrevBLL = &BlockLayoutList;

  unsigned NewAddr;

  cur = ftok.GetNextToken();
  if (cur != "LAYOUT") {
    cur.WriteFileLine(ErrStr);
    ErrStr << "\"LAYOUT\" expected, not \"" << cur << '"' << ENDE;
  }

  do {
    PrevBlock = CurBlock;

    cur = ftok.PeekNextToken();
    if (cur == "ENDL") break;
    if (isdigit(cur.Text[0]) || (cur == '$')) {
      // it's an address;
      NewAddr = ReadWord(ftok);
      if (GP.ProgStart == 0xffff) GP.ProgStart = NewAddr;
      else {
	// fill with zeros up to NewAddr:
	if (NewAddr < GP.ProgStart) {
	  cur.WriteFileLine(ErrStr);
	  ErrStr << "Invalid address" << ENDE;
	}
	if ((NewAddr-GP.ProgStart) < GP.OBufSize) {
	  cur.WriteFileLine(ErrStr);
	  if (PrevBlock)
	    ErrStr << "Code for \"" << PrevBlock->Name
	      << "\" runs over this address" << ENDE;
	  ErrStr << "Invalid address" << ENDE;
	}

	while (GP.OBufSize < (NewAddr-GP.ProgStart))
	  GP.OBuf[GP.OBufSize++] = 0;
      }

      if (ftok.GetNextToken() != ':') {
	cur.WriteFileLine(ErrStr);
	ErrStr << "':' expected" << ENDE;
      }
      continue;
    }

    ftok.GetNextToken();  // read cur

    if (!IsID(cur.Text)) {
      cur.WriteFileLine(ErrStr);
      ErrStr << "Invalid characters in identifier" << ENDE;
    }

    if (GP.ProgStart == 0xffff) {
      cur.WriteFileLine(ErrStr);
      ErrStr << "Missing program start address" << ENDE;
    }

    CurBlock = (TBlock *)BlockList.Find(cur.Text.c_str());

    if (!CurBlock) {
      cur.WriteFileLine(ErrStr);
      ErrStr << "Unknown block name: " << cur.Text << ENDE;
    }

    // Add it to the list:
    CurBLL = new TBlockLayoutList;
    if (!CurBLL) ErrStr.OutOfMemory();
    CurBLL->Block = CurBlock;
    CurBLL->Next = 0;
    *PrevBLL = CurBLL;
    PrevBLL = &(CurBLL->Next);

    // copy the block's object code into dest file:
    CurBlock->OffsetInDestFile = GP.OBufSize;
    unsigned bytecount = 0;
    char *c = CurBlock->ObjCode;
    while (bytecount < CurBlock->ObjCodeSize) {
      GP.OBuf[GP.OBufSize++] = *c;
      bytecount++;
      c++;
    }

  } while(1);
}
