/* 
  SHOWDUMP.CPP - Written by Peter Gonzalez [pbgonz@mail.wm.edu], 1995 
 
    CASM, CDASM, CLINK, and CDUMP are freeware; they may be freely 
  distributed, provided that no fee is charged and I am given due 
  credit for my work.  This source code is available directly from me upon 
  request; please do not distribute the source code or any compiled  
  modifications without my consent.  Support freeware!  
*/ 
 
#include "casm.h"
#include "tables.h"
#include "numbers.h"
#include "block.h"

void disassemble(unsigned &curinstr) {
  // disassembles instruction at curinstr, an offset into GP.OBuf

  // first, write the line number:
  cout << "\n  $";
  WriteHexWord(cout,curinstr + GP.ProgStart);
  cout << ':' << ((char) 9); // tab

  if (instrtable[(unsigned char)(GP.OBuf[curinstr])].instr == QQQ) {
    cout << "dat $";
    WriteHexByte(cout,GP.OBuf[curinstr++]);
    return;
  }

  // next, write the instruction name:
  instrtype InstrCode = instrtable[(unsigned char)(GP.OBuf[curinstr])].instr;
  cout << instrstr[InstrCode] << ' ';

  //now, write the parameters:
  paramtype param = instrtable[(unsigned char)(GP.OBuf[curinstr])].param;

  curinstr++;

  unsigned hex2 = GP.OBuf[curinstr];
  unsigned hex4 = * ((unsigned far *) (GP.OBuf + curinstr));

  switch (param) {
    case TYPEA: // nothing
      switch (InstrCode) {
	case ASL:
	case LSR:
	case ROR:
	case ROL:
	  cout << "acc";
      }

      break;
    case TYPEB: // ($00,X)
      cout << "($"; WriteHexByte(cout,hex2); cout << ",X)";
      curinstr++;
      break;
    case TYPEC: // $00
      cout << '$'; WriteHexByte(cout,hex2);
      curinstr++;
      break;
    case TYPED: // #$00
      cout << "#$"; WriteHexByte(cout,hex2);
      curinstr++;
      break;
    case TYPEE: // ($00),Y
      cout << "($"; WriteHexByte(cout,hex2); cout << "),Y";
      curinstr++;
      break;
    case TYPEF: // $00,X
      cout << '$'; WriteHexByte(cout,hex2); cout << ",X";
      curinstr++;
      break;
    case TYPEL: // $00,Y
      cout << '$'; WriteHexByte(cout,hex2); cout << ",Y";
      curinstr++;
      break;
    case TYPEK: // $0000 for near jumps
      curinstr += 1;
      hex4 = curinstr + GP.ProgStart;
      if (hex2 & 0x80) hex4 += (hex2 & 0x7f) - 0x80;
      else hex4 += (hex2 & 0x7f);
      cout << '$'; WriteHexWord(cout,hex4);
      break;
    case TYPEG: // $0000
      cout << '$'; WriteHexWord(cout,hex4);
      curinstr += 2;
      break;
    case TYPEH: // $0000,Y
      cout << '$'; WriteHexWord(cout,hex4); cout << ",Y";
      curinstr += 2;
      break;
    case TYPEI: // $0000,X
      cout << '$'; WriteHexWord(cout,hex4); cout << ",X";
      curinstr += 2;
      break;
    case TYPEJ: // ($0000)
      cout << "($"; WriteHexWord(cout,hex4); cout << ')';
      curinstr += 2;
      break;
  }
}

void DisassembleBlock(THashObject *cur) {
  if (!cur) return;

  TBlock *Block = (TBlock *)cur;
  if (Block->BlockType == TBlock::DATA) {
    // it's a data block
    cout << "DATA " << Block->Name << " (" << Block->ObjCodeSize
      << " bytes)";

    int curp = Block->OffsetInDestFile;
    int count = Block->ObjCodeSize;
    int wrap = 0;

    while (count--) {
      if (!wrap) {
	cout << "\n  $";
	WriteHexWord(cout,curp + GP.ProgStart);
	cout << ":" << ((char) 9);
      }
      ++wrap %= 8;
      cout << '$';
      WriteHexByte(cout,GP.OBuf[curp++]);
      cout << ' ';
    }
    cout << "\nENDD\n\n";
  } else {
    // it's a data block
    cout << "CODE " << Block->Name << " (" << Block->ObjCodeSize
      << " bytes)";

    unsigned curp = Block->OffsetInDestFile;

    while (curp < (Block->ObjCodeSize + Block->OffsetInDestFile)) {
      disassemble(curp);
    }
    cout << "\nENDC\n\n";
  }
  for (int i=0; i<BTREENUMBRANCHES; i++)
    DisassembleBlock(cur->Branch[i]);
}

const char Divider =
"\n--------------------------------------------------------------------------\n;

void MakeListFile() {

  DString LSTFileName = GP.MainFileName;
  LSTFileName.RemoveFileExtension();
  LSTFileName += ".LST";
  ofstream ListFile(LSTFileName);
  if (ListFile.fail()) {
    cout << "\nERROR: Unable to create " << LSTFileName << '\n';
    return;
  }

  cout << Divider << "List file for " << GP.MainFileName << Divider;

  cout << "\nBlock map:\n";

  cout << "\n\n\n;---------------------------------------------;\n"
       << ";                Disassembly                  ;\n"
       << ";---------------------------------------------;\n";

  cout << ";Load Address: $";
  WriteHexWord(cout,GP.ProgStart);
  if (GP.LeaveOutLoadAddress) cout << " (Not stored in file)";
  cout << "\n\n";

  DisassembleBlock(BlockList.GetRoot());

  cout << ";---------------------------------------------;\n";
}

