/* 
  MISC.CPP - Written by Peter Gonzalez [pbgonz@mail.wm.edu], 1995 
 
    CASM, CDASM, CLINK, and CDUMP are freeware; they may be freely 
  distributed, provided that no fee is charged and I am given due 
  credit for my work.  This source code is available directly from me upon 
  request; please do not distribute the source code or any compiled  
  modifications without my consent.  Support freeware!  
*/ 
 
#include <iostream.h>
#include <ctype.h>

#include "misc.h"
#include "tables.h"
#include "alias.h"
#include "block.h"

BOOL isidchar(int c) {
  return isalpha(c) || (c == '_');
}

//--------------------------------------------------------------------------
BOOL IsID(const FString& str) {
//--------------------------------------------------------------------------
  const char *c = str.c_str();

  if (!isidchar(*c++)) return FALSE;

  while (*c) {
    if (!isidchar(*c) && !isdigit(*c)) return FALSE;
    c++;
  }

  return TRUE;
}


//--------------------------------------------------------------------------
char ASCII2PETASCII(char c) {
//--------------------------------------------------------------------------
  return c;
}

//--------------------------------------------------------------------------
char ASCII2ScreenCode(char c) {
//--------------------------------------------------------------------------
  if ((c >= 'A') && (c <= 'Z')) return c - 'A' + 0x41;
  if ((c >= 'a') && (c <= 'z')) return c - 'a' + 0x01;
  if ((c >= '0') && (c <= '9')) return c - '0' + 0x30;
  switch (c) {
    case '@': return 0;
    case '[': return 0x1B;
    case ']': return 0x1D;
    case '^': return 0x1E;    // up arrow
    case '`': return 0x1F;    // left arrow
    case '\\': return 0x1C;   // lira
    case ' ': return 0x20;
    case '!': return 0x21;
    case '"': return 0x22;
    case '#': return 0x23;
    case '$': return 0x24;
    case '%': return 0x25;
    case '&': return 0x26;
    case '\'': return 0x27;
    case '(': return 0x28;
    case ')': return 0x29;
    case '*': return 0x2A;
    case '+': return 0x2B;
    case ',': return 0x2C;
    case '-': return 0x2D;
    case '.': return 0x2E;
    case '/': return 0x2F;
    case ':': return 0x3A;
    case ';': return 0x3B;
    case '<': return 0x3C;
    case '=': return 0x3D;
    case '>': return 0x3E;
    case '?': return 0x3F;
    case '|': return 0x5D;    // Vertical line
    case '~': return 0x63;    // overscore
    case '_': return 0x64;
  }
  return c;
}

//--------------------------------------------------------------------------
BOOL IsReservedWord(const FString& ID) {
//--------------------------------------------------------------------------
  if ( (ID == "CODE")
    || (ID == "ENDC")
    || (ID == "DATA")
    || (ID == "ENDD")
    || (ID == "VARS")
    || (ID == "ENDV")
    || (ID == "LAYOUT")
    || (ID == "ENDL")
    || (ID == "ALIASW")
    || (ID == "ALIASB")
    || (ID == "VW")
    || (ID == "VB")
    || (ID == "ACC") ) return TRUE;
  return FALSE;
}

//--------------------------------------------------------------------------
BOOL IsNameInUse(const FString& ID) {
//--------------------------------------------------------------------------
  if (!IsID(ID)) return FALSE;  // some other problem

  if (AliasList.Find(ID.c_str())) return TRUE;
  if (BlockList.Find(ID.c_str())) return TRUE;

  if (IsReservedWord(ID)) return TRUE;

  // make sure it's not an instruction:
  if (GetInstrNum(ID.c_str()) >= 0) return TRUE;

  return FALSE;
}

