/* 
  NUMBERS.CPP - Written by Peter Gonzalez [pbgonz@mail.wm.edu], 1995

  IMPORTANT NOTE: This file responds to the "CASM" compiler directive,
		  which should only be defined when compiling CASM.  (See
		  below.)
 
    CASM, CDASM, CLINK, and CDUMP are freeware; they may be freely 
  distributed, provided that no fee is charged and I am given due 
  credit for my work.  This source code is available directly from me upon 
  request; please do not distribute the source code or any compiled  
  modifications without my consent.  Support freeware!  
*/ 
 
#include "numbers.h"
#include "misc.h"

//--------------------------------------------------------------------------
int ReadByteOrWord(TASMFileTokenizer& ftok,unsigned& dat) {
//--------------------------------------------------------------------------
/*
  Attempts to read a hex/dec/bin byte or word into dat.  Returns:
   1	word read
   0	nothing read
  -1	byte read
*/
  unsigned num;
  int digits;
  BOOL Word = FALSE;

  TToken cur = ftok.PeekNextToken();

  if (cur == '$') {
    // hex word
    ftok.GetNextToken(); // read the peeked value

    cur = ftok.GetNextToken();
    if (!HexStr2Int(cur.Text.c_str(),num,digits)) {
      cur.WriteFileLine(ErrStr);
      ErrStr << "Hex number expected" << ENDE;
    }
    if (digits > 2) Word = TRUE;

/*************************************************************************/
// this code is not needed when making CLINK, CDASM, or CDUMP, and is
// omitted because it would require inclusion of other unnecessary modules
#ifdef CASM

  } else if (cur == '%') {
    // bin word
    ftok.GetNextToken(); // read the peeked value

    cur = ftok.GetNextToken();
    const char *c = cur.Text.c_str();

    num = 0;

    // %00000000.11111111
    for (int i=0; i < 17; i++) {
      if (i == 8) {
	if (ftok.PeekNextToken() != '.') {
	  if (*c) {
	    cur.WriteFileLine(ErrStr);
	    ErrStr
	    << "Use \"%00000000.11111111\" format for 16-bit binary" << ENDE;
	  }
	  // it's a byte
	  break;
	}

	Word = TRUE;

	ftok.GetNextToken(); // read the '.'

	cur = ftok.GetNextToken();
	c = cur.Text.c_str();
	continue;
      }

      num <<= 1;
      if (*c == '1') num |= 1;
      else if (*c != '0') {
	cur.WriteFileLine(ErrStr);
	ErrStr << "Invalid binary digit: '" << ((char) *c) << '\'' << ENDE;
      }
      c++;
    }
  } else if (cur == '"') {
    // it's a PETASCII constant:
    ftok.GetNextToken(); // read the peeked value

    num = ASCII2PETASCII(ftok.Get());
    if (ftok.GetNextToken() != '"') {
      quoteerr:;
      cur.WriteFileLine(ErrStr);
      ErrStr << "Only single characters are allowed in quotes in this context"
	<< ENDE;
    }
  } else if (cur == '\'') {
    // it's a screen code constant:
    ftok.GetNextToken(); // read the peeked value

    num = ASCII2ScreenCode(ftok.Get());
    if (ftok.GetNextToken() != '\'')
      goto quoteerr;
#endif
/*************************************************************************/


  } else if (DecStr2Int(cur.Text.c_str(),num,digits)) {
    // dec word
    ftok.GetNextToken(); // read the peeked value

    if ((digits > 3) || (num > 255)) Word = TRUE;

  } else return 0; // not a number

  dat = num;
  return (Word ? 1 : -1);
}

//--------------------------------------------------------------------------
unsigned char ReadByte(TASMFileTokenizer& ftok) {
//--------------------------------------------------------------------------
  unsigned ret;
  TToken err = ftok.PeekNextToken();
  switch (ReadByteOrWord(ftok,ret)) {
    case 1:
      err.WriteFileLine(ErrStr);
      ErrStr << "8-bit (not 16-bit) number expected" << ENDE;
    case 0:
      err.WriteFileLine(ErrStr);
      ErrStr << "8-bit number expected" << ENDE;
  }
  return (unsigned char) ret;
}

//--------------------------------------------------------------------------
unsigned ReadWord(TASMFileTokenizer& ftok) {
//--------------------------------------------------------------------------
  unsigned ret;
  TToken err = ftok.PeekNextToken();
  switch (ReadByteOrWord(ftok,ret)) {
    case -1:
      err.WriteFileLine(ErrStr);
      ErrStr << "16-bit (not 8-bit) number expected" << ENDE;
    case 0:
      err.WriteFileLine(ErrStr);
      ErrStr << "16-bit number expected" << ENDE;
  }
  return ret;
}

//--------------------------------------------------------------------------
void WriteHexByte(ostream& os,unsigned char number) {
//--------------------------------------------------------------------------
  char buf[3];
  int digit;
  int i=2;
  while (i--) {
    digit = number & 15;
    if (digit < 10) buf[i] = '0' + digit;
    else buf[i] = 'A' - 10 + digit;
    number >>= 4;
  }
  buf[2] = 0;
  os << buf;
}


//--------------------------------------------------------------------------
void WriteHexWord(ostream& os,unsigned number) {
//--------------------------------------------------------------------------
  char buf[5];
  int digit;
  int i=4;
  while (i--) {
    digit = number & 15;
    if (digit < 10) buf[i] = '0' + digit;
    else buf[i] = 'A' - 10 + digit;
    number >>= 4;
  }
  buf[4] = 0;
  os << buf;
}

//--------------------------------------------------------------------------
BOOL HexStr2Int(const char *str,unsigned& dest,int& numdigits) {
//--------------------------------------------------------------------------
  dest = 0;
  numdigits = 0;

  const char *c = str;

  while (*c) {
    dest <<= 4;
    if (!isxdigit(*c)) return FALSE;
    if ((*c >= '0') && (*c <= '9')) dest += *c-'0';
    else dest += toupper(*c)-'A' + 10;
    if (++numdigits > 4) return FALSE;
    c++;
  }
  if (!numdigits) return FALSE;
  return TRUE;
}

//--------------------------------------------------------------------------
BOOL DecStr2Int(const char *str,unsigned& dest,int& numdigits) {
//--------------------------------------------------------------------------
  dest = 0;
  numdigits = 0;

  const char *c = str;

  while (*c) {
    dest *= 10;
    if ((*c >= '0') && (*c <= '9')) dest += *c-'0';
    else return FALSE;
    if (++numdigits > 5) return FALSE;
    c++;
  }
  if (!numdigits) return FALSE;
  return TRUE;
}

