/*
  PETESTR.CPP - Written by Peter Gonzalez [pbgonz@mail.wm.edu], 1995 
 
    CASM, CDASM, CLINK, and CDUMP are freeware; they may be freely 
  distributed, provided that no fee is charged and I am given due 
  credit for my work.  This source code is available directly from me upon
  request; pl
  ease do not distribute the source code or any compiled
  modifications without my consent.  Support freeware!  
*/ 
 
#include "petestr.h"

#include <dos.h>
#include <string.h>

int fastcmp(const char far *s1,const char far *s2) {
  return stricmp(s1,s2) == 0;

  // My ultra-fast routine.  Unfortunately, it can't do a case-insensitive
  // compare.
/*
  unsigned s1_seg = FP_SEG(s1);
  unsigned s1_off = FP_OFF(s1);
  unsigned s2_seg = FP_SEG(s2);
  unsigned s2_off = FP_OFF(s2);

  if (s1 == s2) return 1;

  asm {
	// save old registers
	push	ds
	push	es

	// point DS:SI to s1
	mov	ds,s1_seg
	mov	si,s1_off

	// point ES:DI to s2
	mov	es,s2_seg
	mov	di,s2_off

	cld

  } label_1:; asm {
	mov	al,ds:[si]
	or	al,al
	jz	zero

	mov	al,es:[di]
	or	al,al
	jz	zero

	cmpsb
	loope label_1
  }

  notequal:;

  asm {
	// restore regs
	pop	es
	pop	ds
  }

  return 0; // not equal

  zero:;
  asm {
	cmpsb
	jne 	notequal
  }

  asm {
	// restore regs
	pop	es
	pop	ds
  }

  return 1;
*/
}

/******************************************************************
			     Variable-Length
******************************************************************/

DString::DString(const char far *src) {
  _Length = strlen(src);
  delta = 10;
  Str = (char far *) farmalloc(1);
  BufSize = 1;
  if (!Str) ErrStr.OutOfMemory();
  Resize();
  strcpy(Str,src);
}

DString::DString(const FString& src) {
  _Length = src.Length();
  delta = 10;
  Str = (char far *) farmalloc(1);
  BufSize = 1;
  if (!Str) ErrStr.OutOfMemory();
  Resize();
  strcpy(Str,src.c_str());
}

DString::DString(const char c) {
  _Length = 1;
  delta = 10;
  Str = (char far *) farmalloc(1);
  BufSize = 1;
  if (!Str) ErrStr.OutOfMemory();
  Resize();
  Str[0] = c;
  Str[1] = 0;
}
DString::DString(const DString& src) {
  _Length = src._Length;
  delta = src.delta;
  Str = (char far *) farmalloc(1);
  BufSize = 1;
  if (!Str) ErrStr.OutOfMemory();
  Resize();
  strcpy(Str,src.c_str());
}


void DString::RemoveFileExtension() {
  int i = _Length-1;

  while (i > 0) {
    if (Str[i] == '.') {
      Str[i] = 0;
      _Length = i;
      Resize();
      return;
    }
    i--;
  }
}

int DString::NumLines() {
  int lines=1;
  char *c = Str;
  while (*c) {
    if (*c == '\n') lines++;
    c++;
  }
  return lines;
}

ostream& operator << (ostream& os,DString& s) { return os << s.Str; }

void DString::Resize() {
  // resizes buffer to hold a string of size _Length
  int newlen = ((_Length / delta) + 1) * delta + 2;
  if (newlen != BufSize) {
    BufSize = newlen;
    Str = (char far *) farrealloc(Str,BufSize);
    if (!Str) ErrStr.OutOfMemory();
    Str[_Length] = 0;
  }
}

DString& DString::operator =(const DString& src) {
  _Length = src._Length;
  delta = src.delta;
  Resize();
  strcpy(Str,src.Str);
  return *this;
}
DString& DString::operator =(const char far *src) {
  _Length = strlen(src);
  Resize();
  strcpy(Str,src);
  return *this;
}
DString& DString::operator =(const char c) {
  _Length = 1;
  Resize();
  Str[0] = c;
  Str[1] = 0;
  return *this;
}

void DString::ShrinkTo(int MaxLen) {
  // chops chars off end of string until its length <= MaxLen
  if (MaxLen < 0) return;
  if (_Length > MaxLen)  {
    _Length = MaxLen;
    Str[_Length] = 0;
    Resize();
  }
}
void DString::Insert(char c,int pos) {
  if ((pos < 0) || (pos > _Length))
    ErrStr << "Invalid string index" << PROGBUG;
  _Length++;
  Resize();
  // scoot everything up one:
  memmove(Str + pos + 1,Str + pos,(_Length+1) - pos);
  Str[pos] = c;
}
void DString::Delete(int pos) {
  if ((pos < 0) || (pos >= _Length))
    ErrStr << "Invalid string index" << PROGBUG;
  // scoot everything down one:
  memmove(Str + pos,Str + pos + 1,_Length - pos);
  _Length--;
  Resize();
}

DString DString::operator +(const DString& src) {
  DString ret(*this);
  ret += src;
  return ret;
}
DString DString::operator +(const char far *src) {
  DString ret(*this);
  ret += src;
  return ret;
}
DString DString::operator +(const char c) {
  DString ret(*this);
  ret += c;
  return ret;
}

/******************************************************************
			     Fixed-Length
******************************************************************/

void FString::RemoveFileExtension() {
  int i = _Length-1;

  while (i > 0) {
    if (Str[i] == '.') {
      Str[i] = 0;
      _Length = i;
      CheckSize();
      return;
    }
    i--;
  }
}

int FString::NumLines() {
  int lines=1;
  char *c = Str;
  while (*c) {
    if (*c == '\n') lines++;
    c++;
  }
  return lines;
}

void FString::CheckSize() {
  if (_Length >= MAXFIXSTRING) {
    ErrStr << "String length exceeded" << PROGBUG;
  }
}

ostream& operator << (ostream& os,FString& s) { return os << s.Str; }

FString::FString(const char far *src) {
  _Length = strlen(src);
  CheckSize();
  strcpy(Str,src);
}
FString::FString(const DString& src) {
  _Length = src.Length();
  CheckSize();
  strcpy(Str,src.c_str());
}
FString::FString(const char c) {
  _Length = 1;
  CheckSize();
  Str[0] = c;
  Str[1] = 0;
}
FString::FString(const FString& src) {
  _Length = src._Length;
  CheckSize();
  strcpy(Str,src.c_str());
}

FString& FString::operator =(const FString& src) {
  _Length = src._Length;
  CheckSize();
  strcpy(Str,src.Str);
  return *this;
}
FString& FString::operator =(const char far *src) {
  _Length = strlen(src);
  CheckSize();
  strcpy(Str,src);
  return *this;
}
FString& FString::operator =(const char c) {
  _Length = 1;
  CheckSize();
  Str[0] = c;
  Str[1] = 0;
  return *this;
}

void FString::ShrinkTo(int MaxLen) {
  // chops chars off end of string until its length <= MaxLen
  if (_Length > MaxLen)  {
    _Length = MaxLen;
    CheckSize();
    Str[_Length] = 0;
  }
}
void FString::Insert(char c,int pos) {
  if ((pos < 0) || (pos > _Length))
    ErrStr << "Invalid string index" << PROGBUG;
  _Length++;
  CheckSize();
  // scoot everything up one:
  memmove(Str + pos + 1,Str + pos,(_Length+1) - pos);
  Str[pos] = c;
}
void FString::Delete(int pos) {
  if ((pos < 0) || (pos >= _Length))
    ErrStr << "Invalid string index" << PROGBUG;
  // scoot everything down one:
  memmove(Str + pos,Str + pos + 1,_Length - pos);
  _Length--;
  CheckSize();
}

FString FString::operator +(const FString& src) {
  FString ret(*this);
  ret += src;
  return ret;
}
FString FString::operator +(const char far *src) {
  FString ret(*this);
  ret += src;
  return ret;
}
FString FString::operator +(const char c) {
  FString ret(*this);
  ret += c;
  return ret;
}

